"""
Sentence Splitting Operator
Split Text
"""
from relevanceai.operations_new.ops_base import OperationAPIBase
from relevanceai.operations_new.processing.text.sentence_splitting.transform import (
    SentenceSplittingTransform,
)


class SentenceSplitterOps(OperationAPIBase, SentenceSplittingTransform):
    pass



from copy import deepcopy
from typing import List, Dict, Any
from .read_utils import DocReadUtils


class DocWriteUtils(DocReadUtils):
    """This is created as a Mixin for others to easily add to their classes"""

    def run_function_against_all_documents(self, fn, docs, field=None):
        """Run a function against documetns if the field is there"""
        # TODO: Change this to an operator and not a dict comprehension
        if field is not None:
            {fn(self.get_field(field, d)) for d in docs if self.is_field(field, d)}
        else:
            {fn(self.get_field(field, d)) for d in docs}

    @classmethod
    def _is_string_integer(cls, x):
        """Test if a string is numeric"""
        try:
            int(x)
            return True
        except:
            return False

    def set_fields_across_document(self, fields: List[str], doc: Dict, values: List):
        [self.set_field(f, doc, values[i]) for i, f in enumerate(fields)]

    @staticmethod
    def set_field(
        field: str, doc: Dict, value: Any, handle_if_missing=True, inplace=True
    ):
        """
        For nested dictionaries, tries to write to the respective field.
        If you toggle off handle_if_misisng, then it will output errors if the field is
        not found.
        e.g.
        field = kfc.item
        value = "fried wings"
        This should then create the following entries if they dont exist:
        {
            "kfc": {
                "item": "fried wings"
            }
        }
        Args:
            field:
                Field of the document to write.
            doc:
                Python dictionary
            value:
                Value to write
        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> sample_document = {'kfc': {'item': ''}}
            >>> vi_client.set_field('kfc.item', sample_document, 'chickens')
        """
        if not inplace:
            doc = deepcopy(doc)

        fields = field.split(".")
        # Assign a pointer.
        d = doc
        for i, f in enumerate(fields):
            # Assign the value if this is the last entry e.g. stores.fastfood.kfc.item will be item
            if i == len(fields) - 1:
                d[f] = value
            else:
                if f in d.keys():
                    d = d[f]
                else:
                    d.update({f: {}})
                    d = d[f]

        if not inplace:
            return doc

    def set_field_across_documents(
        self, field: str, values: List[Any], docs: List[Dict]
    ):
        """
        For multiple documents, set the right fields.
        e.g.
        field = kfc.item
        value = "fried wings"
        This should then create the following entries if they dont exist:
        {
            "kfc": {
                "item": "fried wings"
            }
        }
        Args:
            field:
                Field of the document to write.
            doc:
                Python dictionary
            value:
                Value to write
        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> sample_document = {'kfc': {'item': ''}}
            >>> vi_client.set_fields('kfc.item', sample_document, 'chickens')
        """
        assert len(values) == len(docs), (
            "Assert that the number of values " + "equates to the number of documents"
        )
        for i, value in enumerate(values):
            self.set_field(field, docs[i], value)



"""
Sklearn Base
"""
from typing import Any, List, Dict, Optional, Union

import numpy as np

from sklearn.base import ClusterMixin

from relevanceai.operations_new.cluster.models.base import ClusterModelBase
from relevanceai.operations_new.cluster.models.sklearn import sklearn_models


class SklearnModel(ClusterModelBase):
    """Sklearn model base"""

    def __init__(
        self,
        model: Union[ClusterMixin, str],
        model_kwargs: dict,
    ):

        if isinstance(model, str):
            assert model in sklearn_models
            model = ClusterModelBase.import_from_string(
                f"sklearn.cluster.{sklearn_models[model]}"
            )
            if model_kwargs is None:
                model_kwargs = {}

            self.model = model(**model_kwargs)
        else:
            # Uncomment out below because this breaks support for sklearn-extra models
            # assert type(model).__name__ in list(sklearn_models.values())
            self.model = model

        super().__init__(model_kwargs=self.model.__dict__)

    def warm_start(self):
        model = SklearnModel.import_from_string("sklearn.cluster.KMeans")
        kwargs = self.model_kwargs
        kwargs["init"] = self.model.init
        self.model = model(kwargs)

    def partial_fit(self, *args, **kwargs) -> Any:
        if hasattr(self.model, "partial_fit"):
            return self.model.partial_fit(*args, **kwargs)
        raise ValueError("Model class does not have a `partial_fit` method")

    def fit_predict(self, vectors, *args, **kwargs) -> List[int]:
        warm_start = kwargs.pop("warm_state", False)
        if warm_start and self.name == "kmeans":
            self.warm_start()

        labels: np.ndarray = self.model.fit_predict(vectors, *args, **kwargs)
        self._centroids = self.calculate_centroids(labels, vectors)
        return labels.tolist()

    def fit(self, *args, **kwargs) -> Any:
        warm_start = kwargs.pop("warm_state", False)
        if warm_start and self.name == "kmeans":
            self.warm_start()

        return self.model.fit(*args, **kwargs)

    def predict(self, vectors, *args, **kwargs) -> List[int]:
        labels: np.ndarray = self.model.predict(vectors, *args, **kwargs)
        self._centroids = self.calculate_centroids(labels, vectors)
        return labels.tolist()



"""
Base Operations Class
Checks fields, contains key abstract methods, etc
"""
from abc import ABC, abstractmethod

from typing import Any, Dict, List, Union

from relevanceai.utils import DocUtils
from relevanceai.client import Credentials

DO_NOT_STORE = ["_centroids", Credentials.__slots__]


class OperationsCheck(ABC, DocUtils):
    def _check_alias(self):
        if self.alias is None:
            raise ValueError("alias is not set. Please supply alias=")
        if self.alias.lower() != self.alias:
            raise ValueError("Alias cannot be lower case.")

    def _check_vector_fields(self):
        """It checks that the vector fields are in the schema, and that they are of the correct type"""
        self._check_vector_field_type()

        if len(self.vector_fields) == 0:
            raise ValueError("No vector fields set. Please supply with vector_fields=")
        self._check_vector_field_names()
        self._check_vector_field_in_schema()

    def _check_vector_field_type(self):
        """If the vector_fields is None, raise an error. If it's a string, force it to be a list"""
        # check the vector fields
        if self.vector_fields is None:
            raise ValueError(
                "No vector_fields has been set. Please supply with vector_fields="
            )
        elif isinstance(self.vector_fields, str):
            # Force it to be a list instead
            self.vector_fields = [self.vector_fields]

    def _check_vector_field_names(self):
        """If the class has a vector_fields attribute, then for each vector field in the vector_fields
        attribute, check that the vector field ends with _vector_. If it doesn't, raise a ValueError

        """
        if hasattr(self, "vector_fields"):
            for vector_field in self.vector_fields:
                if not vector_field.endswith("_vector_"):
                    raise ValueError(
                        "Invalid vector field. Ensure they end in `_vector_`."
                    )

    def _check_vector_field_in_schema(self):
        """> If the dataset has a schema, then the vector field must be in the schema"""
        # TODO: check the schema
        if hasattr(self, "dataset"):
            for vector_field in self.vector_fields:
                if hasattr(self.dataset, "schema"):
                    assert vector_field in self.datasets.schema(self.dataset_id)

    @staticmethod
    def normalize_string(string: str):
        return string.lower().replace("-", "").replace("_", "")

    def _check_fields_in_schema(
        self,
        schema: Dict[str, str],
        fields: List[Union[str, None]],
    ) -> None:
        # Check fields in schema
        if fields is not None:
            for field in fields:
                if field not in schema:
                    raise ValueError(f"{field} not in Dataset schema")


class TransformBase(OperationsCheck):
    """
    To write your own operation, you need to add:
    - name
    - transform
    """

    # Typehints to help with development
    vector_fields: List[str]
    alias: str

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.transform(*args, **kwargs)

    @property
    @abstractmethod
    def name(self) -> str:
        """abstractproperty for name"""
        raise NotImplementedError

    @abstractmethod
    def transform(self, *args, **kwargs) -> List[Dict[str, Any]]:
        """abstractmethod for transform"""
        raise NotImplementedError

    def get_operation_metadata(self, *args, **kwargs) -> Dict[str, Any]:
        """abstractmethod for return metadata for upsertion"""

        values = {k: v for k, v in self.__dict__.items() if k not in DO_NOT_STORE}

        return dict(
            operation=self.name,
            values=values,
        )

    def _get_package_from_model(self, model):
        """Determine the package for a model.
        This can be useful for checking dependencies.
        This may be used across modules for
        deeper integrations

        Parameters
        ----------
        model
            The model to be used for clustering.

        Returns
        -------
            The package name

        """
        # TODO: add support for huggingface integrations
        # such as transformers and sentencetransformers
        model_name = str(model.__class__).lower()
        if "function" in model_name:
            model_name = str(model.__name__)

        if "sklearn" in model_name:
            self.package = "sklearn"

        elif "faiss" in model_name:
            self.package = "faiss"

        elif "hdbscan" in model_name:
            self.package = "hdbscan"

        elif "communitydetection" in model_name:
            self.package = "sentence-transformers"

        else:
            self.package = "custom"
        return self.package

    def _generate_output_field(self, field):
        return f"_{self.name}_.{field.lower().replace(' ', '_')}"

    def get_transformers_device(self, device: int = None):
        """
        Automatically returns a GPU device if there is one. Otherwise,
        returns a CPU device for transformers
        """
        if device is not None:
            return device
        import torch

        if torch.cuda.is_available():
            return 0
        return -1



"""Test labels
"""
import pytest
from relevanceai import Client
from relevanceai import mock_documents
from tests.globals.constants import (
    SAMPLE_DATASET_DATASET_PREFIX,
    generate_random_label,
    generate_random_string,
    generate_random_vector,
    generate_random_integer,
)
import random


def random_vector(vector_length: int = 5):
    return [random.uniform(0, 1) for _ in range(5)]


@pytest.fixture
def test_label_ds(test_client: Client):

    dataset = test_client.Dataset(SAMPLE_DATASET_DATASET_PREFIX + "_label")
    dataset.insert_documents(
        documents=mock_documents(100),
        create_id=True,
    )
    yield dataset
    dataset.delete()


@pytest.fixture
def label_documents():
    return [
        {"label": "value", "label_vector_": random_vector()},
        {"label": "value-2", "label_vector_": random_vector()},
    ]


class TestLabelOps:
    def test_label_ops(self, test_client: Client, label_documents):
        # Add testing for label ops
        from relevanceai.operations_new.label import LabelOps

        ops = LabelOps(
            label_documents=label_documents,
            credentials=test_client.credentials,
            vector_field="sample_1_vector_",
            expanded=True,
        )
        # TODO: Test for expanded = True and False
        # Test for differnet label documents (is an intuitive error returned)
        documents = mock_documents(100)
        docs = ops.transform(
            documents=documents,
        )
        assert True



"""A cache mixin
"""


def _is_cache_function(func):
    """Determines if a function is cached or not"""
    return hasattr(func, "cache_info")


def rdir_cache_functions(base_function, function_store: set = None):
    """Recursively gets all the cached functions"""
    if function_store is None:
        function_store = set()
    all_attrs = dir(base_function)
    for attr in all_attrs:
        func = getattr(base_function, attr)
        # Check if it is a class but not that it is a global variable
        # class is denoted by capitalised first letter
        # global variable is denoted by all upper case
        res = _is_cache_function(func)
        if attr[0].isupper() and not attr.upper() == attr:
            rdir_cache_functions(func, function_store=function_store)
        # check for internal classes
        elif attr[0] == "_" and attr[1].isupper() and not attr.upper() == attr:
            rdir_cache_functions(func, function_store=function_store)
        elif res:
            function_store.add(func)
    return function_store


class CacheMixin:
    def _get_all_cache_functions(self):
        import relevanceai

        cache_functions = set()
        for directory in [
            relevanceai.client.client,
            relevanceai.client.client.Dataset,
            relevanceai.dataset.series.Series,
        ]:
            cache_functions.update(rdir_cache_functions(directory))
        return cache_functions

    def clear_cache(self, *args, **kw):
        """
        Clears cache.

        .. code-block::

            from relevanceai import Client
            client = Client()
            client.cache_info()

        """
        cache_functions = self._get_all_cache_functions()
        for func in cache_functions:
            if hasattr(func, "clear_cache"):
                func.clear_cache(*args, **kw)

    def cache_info(self):
        """
        Prints statements explaining how much is stored in each cache.

        Example
        ---------

        .. code-block::

            from relevanceai import Client
            client = Client()
            client.cache_info()

        """
        cache_functions = self._get_all_cache_functions()
        for func in cache_functions:
            if hasattr(func, "cache_info"):
                print(func)
                info = func.cache_info()
                print(info)



"""
The base for a partial cluster model base.
"""
from abc import abstractmethod, ABC
from relevanceai.operations_new.cluster.models.base import _ModelUtils


class BatchClusterModelBase(_ModelUtils, ABC):
    @abstractmethod
    def partial_fit(self):
        raise NotImplementedError

    @abstractmethod
    def predict(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def name(self):
        raise NotImplementedError



"""
Labelling with API-related functions
"""
from relevanceai.operations_new.label.transform import LabelTransform
from relevanceai.operations_new.ops_base import OperationAPIBase


class LabelOps(LabelTransform, OperationAPIBase):  # type: ignore
    """
    Label Operations
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import numpy as np
from relevanceai.operations.dr.base import DimReductionBase
from typing import Optional, Dict, Any
from relevanceai.operations.cluster.constants import (
    DIM_REDUCTION,
    DIM_REDUCTION_DEFAULT_ARGS,
)


class PCA(DimReductionBase):
    def fit(self, vectors: np.ndarray, dims: int = 3, *args, **kw):
        from sklearn.decomposition import PCA as SKLEARN_PCA

        pca = SKLEARN_PCA(n_components=min(dims, vectors.shape[1]))
        return pca.fit(vectors)

    def fit_transform(
        self,
        vectors: np.ndarray,
        dr_args: Optional[Dict[Any, Any]] = DIM_REDUCTION_DEFAULT_ARGS["pca"],
        dims: int = 3,
    ) -> np.ndarray:
        from sklearn.decomposition import PCA as SKLEARN_PCA

        self.logger.debug(f"{dr_args}")
        vector_length = len(vectors[0])
        pca = SKLEARN_PCA(n_components=min(dims, vector_length), **dr_args)
        return pca.fit_transform(vectors)



# -*- coding: utf-8 -*-
"""Batch client to allow for batch insertions/retrieval and encoding
"""
from typing import Callable

from relevanceai._api.batch.insert import BatchInsertClient


class APIClient(BatchInsertClient):
    """Batch API client"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def batch_insert(self):
        raise NotImplemented

    def batch_get_and_edit(self, dataset_id: str, chunk_size: int, bulk_edit: Callable):
        """Batch get the documents and return the documents"""
        raise NotImplemented

    # Other useful utilities
    def append_metadata_list(
        self,
        field: str,
        value_to_append,
        only_unique: bool = False,
        verbose: bool = True,
    ):
        """
        Easily append to a list value in metadata using this function.
        This prevents overwriting if there is one.
        """
        metadata = self.datasets.metadata(self.dataset_id)["results"]
        if self.is_field(field, metadata):
            value_list = self.get_field(field, metadata)
        else:
            value_list = []

        if not isinstance(value_list, list):
            raise ValueError("Can't append to object that is not a list.")

        # Used for not including multiple entries when subclustering
        if only_unique:
            for v in value_list:
                if v == value_to_append:
                    return

        value_list.append(value_to_append)
        self.set_field(field, metadata, value_list)
        self.datasets.post_metadata(
            dataset_id=self.dataset_id,
            metadata=metadata,
        )



"""
.. warning::

    This is a beta feature and will be changing in the future. Do not use this in production systems.

Example
---------------------------------------

Train a text model using TripletLoss

You can find out more about different types of TripletLoss on https://www.sbert.net/docs/package_reference/losses.html

.. code-block::

    from relevanceai import Client
    client = Client()
    ds = client.Dataset("ecommerce")
    ops = SupervisedTripleLossFinetuneOps.from_dataset(
        dataset=ds,
        base_model="distilbert-base-uncased",
        chunksize=16,
        triple_loss_type:str='BatchHardSoftMarginTripletLoss'
    )
    ops.run(text_field="detail_desc", label_field="_cluster_.desc_use_vector_.kmeans-10", output_dir)

"""

import os
import json
import random
import copy
import logging
from typing import List, Optional, Any

try:
    from sentence_transformers import (
        SentenceTransformer,
        SentencesDataset,
        losses,
        evaluation,
    )
    from sentence_transformers.readers import InputExample
except ModuleNotFoundError:
    print(
        "Need to install SentenceTransformer by running `pip install sentence_transformers`."
    )

try:
    from torch.utils.data import DataLoader
except ModuleNotFoundError:
    print("Need to install SentenceTransformer by running `pip install torch-utils`.")

from relevanceai.operations.base import BaseOps
from relevanceai._api.api_client import APIClient
from relevanceai.client.helpers import Credentials


class SupervisedTripleLossFinetuneOps(APIClient, BaseOps):
    def __init__(
        self,
        dataset,
        base_model: str = "sentence-transformers/all-mpnet-base-v2",
        triple_loss_type: str = "BatchHardSoftMarginTripletLoss",
        chunksize: int = 32,
        save_best_model: bool = True,
        credentials: Optional[Credentials] = None,
    ):
        self.dataset = dataset
        self.base_model = base_model
        self.model = SentenceTransformer(base_model)
        self.chunksize = chunksize
        self.save_best_model = save_best_model
        self.loss_type = triple_loss_type
        self.evaluator = None

        if credentials is not None:
            super().__init__(credentials)

    def define_loss(self):
        if self.loss_type == "BatchAllTripletLoss":
            self.loss = losses.BatchAllTripletLoss(model=self.model)
        elif self.loss_type == "BatchHardSoftMarginTripletLoss":
            self.loss = losses.BatchHardSoftMarginTripletLoss(model=self.model)
        elif self.loss_type == "BatchHardTripletLoss":
            self.loss = losses.BatchHardTripletLoss(model=self.model)
        else:
            self.loss = losses.BatchSemiHardTripletLoss(model=self.model)

    def define_evaluator(
        self,
        text_data: List[str],
        labels: List[int],
        name="supervised_finetune_dev_eval",
    ):
        anchors, positives, negatives = self.build_triple_data(text_data, labels)
        self.evaluator = evaluation.TripletEvaluator(
            anchors=anchors, positives=positives, negatives=negatives, name=name
        )

    @staticmethod
    def build_triple_data(text_data, labels):
        # Only used for evaluation
        # anchors: Sentences to check similarity to
        # positives: List of positive sentences
        # negatives:  List of negative sentences

        n = 2  # ToDo: add checks, so that n can be an argument
        label_data_dict = {l: set() for l in set(labels)}
        for d, l in zip(text_data, labels):
            label_data_dict[l].add(d)
        label_data_dict = {l: list(label_data_dict[l]) for l in label_data_dict}
        all_cls = copy.deepcopy(label_data_dict)
        all_cls = list(all_cls.keys())
        for l in all_cls:
            if len(label_data_dict[l]) < 2 * n + 1:
                del label_data_dict[l]

        anchors = []
        positives = []
        negatives = []
        for l in label_data_dict:
            neg_list = list(label_data_dict.keys())
            neg_list.remove(l)
            anchors.extend(label_data_dict[l][:n])
            for idx in random.sample([j for j in range(n, len(label_data_dict[l]))], n):
                positives.append(label_data_dict[l][idx])
            for nl in random.sample(neg_list, n):
                idx = random.choice([j for j in range(n, len(label_data_dict[nl]))])
                negatives.append(label_data_dict[nl][idx])
        return anchors, positives, negatives

    def prepare_data_for_finetuning(self, text_data: List[str], labels: List[int]):
        data = [
            InputExample(texts=[text], label=label)
            for text, label in zip(text_data, labels)
        ]
        dataset = SentencesDataset(data, self.model)
        return DataLoader(dataset, shuffle=True, batch_size=self.chunksize)

    def fine_tune(
        self,
        train_data: List,
        dev_data: List = None,
        epochs: int = 3,
        output_path: str = "trained_model",
    ):
        if os.path.exists(output_path):
            print(
                "Output directory is detected. Assuming model was trained. Change output directory if you want to train a new model."
            )
        # Generates pairs for fine-tuning a model following the GPL algorithm. The final model is saved at  output_dir
        self.path_to_finetuned_model = output_path
        self.output_path = output_path
        print(f"Finetuning {self.base_model}")
        train_set, train_labels = train_data
        self.define_loss()
        dataloader = self.prepare_data_for_finetuning(train_set, train_labels)
        objectives = [(dataloader, self.loss)]
        if self.evaluator and dev_data:
            dev_set, dev_labels = dev_data
            self.define_evaluator(dev_set, dev_labels)
        self.model.fit(
            train_objectives=objectives,
            epochs=epochs,
            evaluator=self.evaluator,
            save_best_model=self.save_best_model,
            output_path=self.output_path,
        )
        print(f"Finished finetuning. Trained model is saved at {self.output_path}")

    def get_model(self, output_path: Optional[str] = None):
        if not self.output_path and not output_path:
            logging.warning("No Fine-Tuned Model Was Found")
        elif self.output_path:
            return SentenceTransformer(self.output_path)
        elif output_path:
            return SentenceTransformer(self.output_path)

    def fetch_text_and_labels_from_dataset(self, text_field, label_field):
        print("Fetching documents...")
        docs = self.dataset.get_all_documents(select_fields=[text_field, label_field])
        text_data = [doc[text_field] for doc in docs]
        labels = [self.client.get_field(label_field, doc) for doc in docs]
        labels_cls = set(labels)
        label_maps = {cl: i for i, cl in enumerate(labels_cls)}
        labels = [label_maps[l] for l in labels]
        return text_data, labels

    def run(
        self,
        text_field: str,
        label_field: str,
        epochs: int = 3,
        output_dir: str = "trained_model",
        percentage_for_dev: float = None,
    ):
        """
        Supervised finetuning a model using TripleLoss

        Example
        -----------

        .. code-block::

            from relevanceai import Client
            client = Client()

            ds = client.Dataset("quickstart")
            from relevanceai.operations.text_finetuning.supervised_finetuning_ops import SupervisedTripleLossFinetuneOps
            ops = SupervisedTripleLossFinetuneOps.from_dataset(ds)
            ops.run(text_field="detail_desc", label_field="_cluster_.desc_use_vector_.kmeans-10", output_dir)

        Parameters
        -------------

        text_field: str
            The field you want to use as input text for fine-tuning
        label_field: str
            The field indicating the classes of the input
        output_dir: str
            The path of the output directory
        percentage_for_dev: float
            a number between 0 and 1 showing how much of the data should be used for evaluation. No evaluation if None

        """

        text_data, labels = self.fetch_text_and_labels_from_dataset(
            text_field, label_field
        )

        if percentage_for_dev:
            trin_size = int(len(labels) * (1 - percentage_for_dev))
            train_data = [text_data[:trin_size], labels[:trin_size]]
            dev_data = [text_data[trin_size:], labels[trin_size:]]
            self.do_evaluation = True
        else:
            train_data = [text_data, labels]
            dev_data = None
        self.fine_tune(train_data, dev_data, epochs, output_dir)

    @classmethod
    def from_client(self, client, *args, **kwargs):
        credentials = client.credentials
        return self(
            credentials=credentials,
            *args,
            **kwargs,
        )

    @classmethod
    def from_dataset(
        self,
        dataset: Any,
        base_model: str = "distilbert-base-uncased",
        **kwargs,
    ):
        cls = self(
            credentials=dataset.credentials,
            base_model=base_model,
            **kwargs,
        )
        cls.dataset_id = dataset.dataset_id
        return cls



"""
.. warning::

    This is a beta feature and will be changing in the future. Do not use this in production systems.

Example
---------------------------------------

Train a text model using GPL (Generative Pseudo-Labelling)
This can be helpful for `domain adaptation`.

You can find out more about GPL from: https://github.com/UKPLab/gpl

.. code-block::

    from relevanceai import Client
    client = Client()
    ds = client.Dataset("ecommerce")
    ops = GPLOps.from_dataset(dataset=ds,
        base_model="distilbert-base-uncased",
        t5_generator="BeIR/query-gen-msmarco-t5-base-v1",
        retrievers=["msmarco-distilbert-base-v3", "msmarco-MiniLM-L-6-v3"],
        cross_encoder="cross-encoder/ms-marco-MiniLM-L-6-v2",
        chunksize_gpl=16,
        output_path="trained_model",
    )
    ops.run(dataset=ds, text_field="detail_desc")

"""

import os
import json
import logging
from typing import List, Optional, Any

try:
    import gpl
except ModuleNotFoundError:
    print("Need to install gpl by running `pip install gpl`.")

try:
    from sentence_transformers import SentenceTransformer
except ModuleNotFoundError:
    print(
        "Need to install SentenceTransformer by running `pip install sentence_transformers`."
    )
from relevanceai.operations.base import BaseOps
from relevanceai._api.api_client import APIClient
from relevanceai.client.helpers import Credentials


class GPLOps(APIClient, BaseOps):
    def __init__(
        self,
        base_model: str = "distilbert-base-uncased",
        t5_generator: str = "BeIR/query-gen-msmarco-t5-base-v1",
        retrievers: List[str] = ["msmarco-distilbert-base-v3", "msmarco-MiniLM-L-6-v3"],
        cross_encoder: str = "cross-encoder/ms-marco-MiniLM-L-6-v2",
        chunksize_gpl: int = 16,
        output_path: str = "trained_model",
        credentials: Optional[Credentials] = None,
    ):
        self.base_model = base_model
        self.t5_generator = t5_generator
        self.retrievers = retrievers
        self.cross_encoder = cross_encoder
        self.chunksize_gpl = chunksize_gpl
        self.output_path = output_path
        if credentials is not None:
            super().__init__(credentials)

    def prepare_data_for_finetuning(
        self,
        documents: List[dict],
        text_field: str,
        dir_to_save_corpus: str,
        title_field: Optional[str] = None,
        corpus_filename: str = "corpus.jsonl",
    ):
        # Writes a corpus in the format compatible to the GPL library to later make Pos/Neg pairs
        corpus = []
        i = 0
        with open(f"{dir_to_save_corpus}/{corpus_filename}", "w") as jsonl:
            for doc in documents:
                if text_field not in doc:
                    continue
                line = {
                    "_id": str(i),
                    "text": doc.get(text_field, "").replace("\n", " "),
                    "metadata": {
                        f: doc[f] for f in doc if f not in [text_field, title_field]
                    },
                }
                if title_field is not None:
                    line["title"] = doc.get(title_field, "")
                else:
                    line["title"] = ""
                # iteratively write lines to the JSON lines corpus.jsonl file
                jsonl.write(json.dumps(line) + "\n")
                i += 1
                corpus.append(line)
        logging.info(
            f"A corpus of {len(corpus)} documents is saved at {dir_to_save_corpus}/corpus.jsonl"
        )

    def fine_tune(
        self,
        path_to_generated_data: str = ".",
        output_dir: str = "trained_model",
        gpl_steps: int = 500,
        do_evaluation: bool = False,
        qgen_prefix: str = "qgen",
        **gpl_kwargs,
    ):
        if os.path.exists(output_dir):
            print(
                "Output directory is detected. Assuming model was trained. Change output directory if you want to train a new model."
            )
        # Generates pairs for fine-tuning a model following the GPL algorithm. The final model is saved at  output_dir
        self.path_to_finetuned_model = output_dir
        self.output_path = output_dir
        # TODO: Add the rest of the parameters :(
        gpl.train(
            path_to_generated_data=path_to_generated_data,
            base_ckpt=self.base_model,
            batch_size_gpl=self.chunksize_gpl,
            gpl_steps=gpl_steps,
            output_dir=output_dir,
            generator=self.t5_generator,
            retrievers=self.retrievers,
            cross_encoder=self.cross_encoder,
            qgen_prefix=qgen_prefix,
            do_evaluation=do_evaluation,
            **gpl_kwargs,
        )

    def get_model(self, output_path: Optional[str] = None):
        if not self.output_path and not output_path:
            logging.warning("No Fine-Tuned Model Was Found")
        elif self.output_path:
            return SentenceTransformer(self.output_path)
        elif output_path:
            return SentenceTransformer(self.output_path)

    def run(
        self,
        dataset: str,
        text_field: str,
        gpl_steps: int = 500,
        path_to_generated_data: str = ".",
        output_dir: str = "trained_model",
        dir_to_save_corpus: str = ".",
        do_evaluation: bool = False,
    ):
        """
        Finetune a model using Generative Pseudo-Labelling

        Example
        -----------

        .. code-block::

            from relevanceai import Client
            client = Client()

            ds = client.Dataset("quickstart")
            # !pip install -q gpl
            from relevanceai.operations.text_finetuning.unsupervised_finetuning_ops import GPLOps
            ops = GPLOps.from_dataset(ds)
            ops.run(dataset=ds, text_field="product_title")

        Parameters
        -------------

        text_field: str
            The field you want to use for fine-tuning
        dir_to_save_corpus: str
            path to save the corpus that is going to be used by the GPL alg.
        gpl_steps: int
            The number of steps in Generative Pseudo Labelling
        path_to_generated_data: str
            The path to generated data
        output_dir: str
            The path of the output directory
        dir_to_save_corpus: str
            The directory to save corpus
        do_evaluation: bool
            If True, it performs the evaluation

        """
        print("Fetching documents...")
        docs = self._get_all_documents(
            dataset_id=self._get_dataset_id(dataset), select_fields=[text_field]
        )

        print("Preparing training data...")
        self.prepare_data_for_finetuning(
            documents=docs,
            text_field=text_field,
            dir_to_save_corpus=dir_to_save_corpus,
        )

        print("Training Model...")
        self.fine_tune(
            path_to_generated_data=path_to_generated_data,
            output_dir=output_dir,
            gpl_steps=gpl_steps,
            do_evaluation=do_evaluation,
        )

        print(f"Trained model saved at {self.output_path}")

    @classmethod
    def from_client(self, client, *args, **kwargs):
        credentials = client.credentials
        return self(
            credentials=credentials,
            *args,
            **kwargs,
        )

    @classmethod
    def from_dataset(
        self,
        dataset: Any,
        base_model: str = "distilbert-base-uncased",
        t5_generator: str = "BeIR/query-gen-msmarco-t5-base-v1",
        retrievers: List[str] = ["msmarco-distilbert-base-v3", "msmarco-MiniLM-L-6-v3"],
        cross_encoder: str = "cross-encoder/ms-marco-MiniLM-L-6-v2",
        chunksize_gpl: int = 16,
        output_path: str = "trained_model",
        **kwargs,
    ):
        cls = self(
            credentials=dataset.credentials,
            base_model=base_model,
            t5_generator=t5_generator,
            retrievers=retrievers,
            cross_encoder=cross_encoder,
            chunksize_gpl=chunksize_gpl,
            output_path=output_path,
            **kwargs,
        )
        cls.dataset_id = dataset.dataset_id
        return cls



"""Cosine similarity operations
"""
from typing import List, Dict, Any

import numpy as np

from relevanceai.utils.integration_checks import is_scipy_available
from relevanceai.utils.decorators.analytics import track
from relevanceai.utils import DocUtils


@track
def cosine_similarity_matrix(a, b, decimal=None):
    A = np.array(a)
    B = np.array(b)
    similarity = np.dot(A, B.T)
    square_mag = np.diag(similarity)
    inv_square_mag = 1 / square_mag
    inv_square_mag[np.isinf(inv_square_mag)] = 0
    inv_mag = np.sqrt(inv_square_mag)
    cosine = similarity * inv_mag
    cosine = cosine.T * inv_mag
    cosine[cosine > 0.9999] = 1
    if decimal:
        cosine = np.around(cosine, decimal)
    return cosine.tolist()


@track
def cosine_similarity(a, b):
    """Cosine similarity utility"""
    if is_scipy_available():
        from scipy import spatial

        return 1 - spatial.distance.cosine(a, b)
    else:
        a_array = np.array(a)
        b_array = np.array(b)
        return a_array.dot(b_array) / (
            np.linalg.norm(a_array, axis=1) * np.linalg.norm(b_array)
        )


def get_cosine_similarity_scores(
    self,
    documents: List[Dict[str, Any]],
    anchor_document: Dict[str, Any],
    vector_field: str,
) -> List[float]:
    """
    Compare scores based on cosine similarity

    Args:
        other_documents:
            List of documents (Python Dictionaries)
        anchor_document:
            Document to compare all the other documents with.
        vector_field:
            The field in the documents to compare
    Example:
        >>> documents = [{...}]
        >>> ViClient.get_cosine_similarity_scores(documents[1:10], documents[0])

    """
    similarity_scores = []
    for i, doc in enumerate(documents):
        similarity_score = self.calculate_cosine_similarity(
            self.get_field(vector_field, doc),
            self.get_field(vector_field, anchor_document),
        )
        similarity_scores.append(similarity_score)
    return similarity_scores


def largest_indices(
    ary,
    n,
):
    """
    Returns the n largest indices from a numpy array.

    Code from: https://stackoverflow.com/questions/6910641/how-do-i-get-indices-of-n-maximum-values-in-a-numpy-array

    """
    flat = ary.flatten()
    indices = np.argpartition(flat, -n)[-n:]
    indices = indices[np.argsort(-flat[indices])]
    return np.unravel_index(indices, ary.shape)



from relevanceai import Client
from relevanceai import ClusterOps
from relevanceai.dataset import Dataset

# Comment out these tests for now as they are not implemented

# class TestBaseOps:
#     def test_init(self, test_client: Client):
#         credentials = test_client.credentials
#         ops = ClusterOps.init(credentials=credentials)
#         assert True

#     def test_from_credentials(self, test_client: Client):
#         credentials = test_client.credentials
#         ops = ClusterOps.init(credentials=credentials)
#         assert ops.credentials.token == credentials.token

#     def test_from_token(self, test_token: str):
#         ops = ClusterOps.from_token(token=test_token)
#         assert ops.credentials.token == test_token

#     def test_from_client(self, test_client: Client):
#         ops = ClusterOps.from_client(client=test_client)
#         assert ops.credentials.token == test_client.credentials.token

#     def test_from_dataset(self, test_clustered_df: Dataset):
#         ops = ClusterOps.from_dataset(
#             dataset=test_clustered_df,
#             alias="kmeans-10",
#             vector_fields=["sample_1_vector_"],
#         )
#         closest = ops.list_closest()
#         assert closest



import numpy as np
from typing import Optional, List, Any, Union, Dict, Tuple, Set
from tqdm.auto import tqdm
from relevanceai.operations.cluster.partial import PartialClusterOps
from relevanceai.operations.cluster.cluster import ClusterOps
from relevanceai._api import APIClient
from relevanceai.operations.cluster.utils import ClusterUtils


class _SubClusterOps(ClusterOps):
    """This class is an intermediate layer between cluster ops and subclusterops.
    It it used to over-write parts of clusterops.
    """

    def __init__(
        self,
        credentials,
        alias: str,
        model,
        dataset: Any,
        vector_fields: list,
        parent_field: str,
        outlier_value=-1,
        outlier_label="outlier",
        verbose=False,
        **kwargs,
    ):
        """
        Sub Cluster Ops
        """

        # self.dataset = dataset
        self.vector_fields = vector_fields
        self.parent_field = parent_field
        self.credentials = credentials
        self.model = model
        self.outlier_value = outlier_value
        self.outlier_label = outlier_label
        self.verbose = verbose
        if isinstance(dataset, str):
            self.dataset_id = dataset
        else:
            if hasattr(dataset, "dataset_id"):
                self.dataset_id = dataset.dataset_id
            # needs to have the dataset_id attribute

        if "package" not in self.__dict__:
            self.package = self._get_package(self.model)

        self.model_name = None
        self.model = self._get_model(model)
        self.alias = self._get_alias(alias)
        super().__init__(
            credentials=credentials,
            model=self.model,
            vector_fields=self.vector_fields,
            alias=self.alias,
            dataset_id=self.dataset_id,
            outlier_value=self.outlier_value,
            outlier_label=self.outlier_label,
        )

    def run(
        self,
        dataset_id: Optional[Union[str, Any]] = None,
        parent_field: str = None,
        vector_fields: Optional[List[str]] = None,
        filters: list = None,
        show_progress_bar: bool = True,
        verbose: bool = True,
        min_parent_cluster_size: Optional[int] = None,
    ) -> None:
        """
        Run clustering on a dataset

        Parameters
        --------------

        dataset_id: Optional[Union[str, Any]]
            The dataset ID
        vector_fields: Optional[List[str]]
            List of vector fields
        show_progress_bar: bool
            If True, the progress bar can be shown

        """

        if parent_field is None:
            parent_field = self.parent_field

        filters = [] if filters is None else filters

        if not isinstance(dataset_id, str):
            if hasattr(dataset_id, "dataset_id"):
                dataset_id = dataset_id.dataset_id  # type: ignore

        if vector_fields is None:
            vector_fields = self.vector_fields

        self.dataset_id = dataset_id
        if vector_fields is not None:
            vector_field = vector_fields[0]
            self.vector_field = vector_field

        # get all documents
        documents = self._get_all_documents(
            dataset_id=dataset_id,
            select_fields=vector_fields + [parent_field],
            show_progress_bar=show_progress_bar,
            include_vector=True,
            filters=filters,
        )

        # If no documents then return
        if len(documents) == 0:
            return

        if (
            min_parent_cluster_size is not None
            and len(documents) < min_parent_cluster_size
        ):
            return

        # fit model, predict and label all documents
        centroid_documents, labelled_documents = self._fit_predict(
            documents=documents,
            vector_field=vector_field,
        )

        results = self._update_documents(
            dataset_id=dataset_id,
            documents=labelled_documents,
            show_progress_bar=show_progress_bar,
        )

        self._insert_centroids(
            dataset_id=dataset_id,
            vector_field=vector_field,
            centroid_documents=centroid_documents,
        )

        # link back to dashboard
        if verbose:
            self._print_app_link()

    def _format_sub_labels(self, parent_values: list, labels: np.ndarray) -> List[str]:
        if len(parent_values) != len(labels):
            raise ValueError("Improper logic for parent values")

        labels = labels.flatten().tolist()
        cluster_labels = [
            label + "-" + str(labels[i])
            if label != self.outlier_value
            else self.outlier_label
            for i, label in enumerate(parent_values)
        ]
        return cluster_labels

    def _get_centroid_documents(
        self, vectors: np.ndarray, labels: List[str]
    ) -> List[Dict[str, Any]]:
        centroid_documents = []

        centroids: Dict[str, Any] = {}
        for label, vector in zip(labels, vectors.tolist()):
            if label not in centroids:
                centroids[label] = []
            centroids[label].append(vector)

        for centroid, vectors in centroids.items():
            centroid_vector = np.array(vectors).mean(0).tolist()
            centroid_document = dict(
                _id=centroid,
                centroid_vector=centroid_vector,
            )
            centroid_documents.append(centroid_document)

        return centroid_documents

    def _insert_centroids(
        self,
        dataset_id: str,
        vector_field: str,
        centroid_documents: List[Dict[str, Any]],
    ) -> None:
        self.datasets.cluster.centroids.insert(
            dataset_id=dataset_id,
            cluster_centers=centroid_documents,
            vector_fields=[vector_field],
            alias=self.alias,
        )

    def _fit_predict(
        self,
        documents: List[Dict[str, Any]],
        vector_field: str,
        return_only_labels=True,
    ) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
        doc_subset = [doc for doc in documents if self.is_field(vector_field, doc)]
        vectors = np.array(
            [self.get_field(vector_field, document) for document in doc_subset]
        )

        if self.package == "sklearn":
            labels = self.model.fit_predict(vectors)

        elif self.package == "hdbscan":
            self.model.fit(vectors)
            labels = self.model.labels_.reshape(-1, 1)

        elif self.package == "faiss":
            vectors = vectors.astype("float32")
            self.model.train(vectors)
            labels = self.model.assign(vectors)[1]

        elif self.package == "sentencetransformers":
            # TODO: make teh config here better
            labels = self.model(vectors)
            labels = np.array(labels)

        elif self.package == "custom":
            labels = self.model.fit_predict(vectors)
            if not isinstance(labels, np.ndarray):
                raise ValueError("Custom model must output np.ndarray for labels")

            if labels.shape[0] != vectors.shape[0]:
                raise ValueError("incorrect shape")

        parent_values = self.get_field_across_documents(self.parent_field, doc_subset)
        labels = self._format_sub_labels(parent_values, labels)

        cluster_field = f"_cluster_.{vector_field}.{self.alias}"
        if return_only_labels:
            doc_subset = [
                {"_id": doc["_id"], cluster_field: labels[i]}
                for i, doc in enumerate(doc_subset)
            ]
        else:
            self.set_field_across_documents(
                field=cluster_field, values=labels, docs=doc_subset
            )

        centroid_documents = self._get_centroid_documents(vectors, labels)

        return centroid_documents, doc_subset


class SubClusterOps(_SubClusterOps, ClusterUtils):  # type: ignore
    def __init__(
        self,
        credentials,
        alias: str,
        dataset,
        model,
        vector_fields: List[str],
        parent_field: str,
        outlier_value=-1,
        outlier_label="outlier",
        verbose: bool = True,
        **kwargs,
    ):
        """
        Sub Cluster Ops
        """
        self.alias = alias
        # self.dataset = dataset
        self.vector_fields = vector_fields
        self.parent_field = parent_field
        self.credentials = credentials
        self.model = model
        self.outlier_value = outlier_value
        self.outlier_label = outlier_label
        self.verbose = verbose
        if isinstance(dataset, str):
            self.dataset_id: str = dataset
        else:
            if hasattr(dataset, "dataset_id"):
                self.dataset_id = dataset.dataset_id
            # needs to have the dataset_id attribute
        self.n_clusters = kwargs.get("n_clusters", 4)
        self.cluster_config = kwargs.get("cluster_config", {})

        super().__init__(
            credentials=credentials,
            alias=alias,
            model=model,
            dataset=dataset,
            vector_fields=vector_fields,
            parent_field=parent_field,
        )

    def __call__(self, *args, **kw):
        return self.fit_predict(*args, **kw)

    def fit_predict(
        self,
        dataset,
        vector_fields: List[Any],
        parent_field: str = None,
        filters: Optional[List] = None,
        verbose: bool = False,
        min_parent_cluster_size: Optional[int] = None,
        cluster_ids: Optional[list] = None,
    ):
        """

        Run subclustering on your dataset using an in-memory clustering algorithm.

        Parameters
        --------------

        dataset: Dataset
            The dataset to create
        vector_fields: List
            The list of vector fields to run fitting, prediction and updating on
        filters: Optional[List]
            The list of filters to run clustering on
        verbose: bool
            If True, this should be verbose

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()

            from relevanceai.package_utils.datasets import mock_documents
            ds = client.Dataset("sample")

            # Creates 100 sample documents
            documents = mock_documents(100)
            ds.upsert_documents(documents)

            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=10)
            clusterer = ClusterOps(alias="minibatchkmeans-10", model=model)
            clusterer.subcluster_predict_update(
                dataset=ds,
            )

        """
        if self.model is None:
            raise ValueError("No model is detected.")

        filters = [] if filters is None else filters

        # load the documents
        self.logger.warning(
            "Retrieving documents... This can take a while if the dataset is large."
        )

        # self.parent_field = parent_field

        # self._init_dataset(dataset)
        self.vector_fields = vector_fields  # type: ignore

        # make sure to only get fields where vector fields exist
        filters += [
            {
                "field": f,
                "filter_type": "exists",
                "condition": "==",
                "condition_value": " ",
            }
            for f in vector_fields  # type: ignore
        ]

        if verbose:
            print("Fitting and predicting on all documents")
        # Here we run subfitting on these documents

        clustered_docs = self.subcluster_predict_documents(
            vector_fields=vector_fields,
            filters=filters,
            verbose=False,
            min_parent_cluster_size=min_parent_cluster_size,
            cluster_ids=cluster_ids,
        )

        if verbose:
            print(
                "Build your clustering app here: "
                + f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
            )

        # Store subcluster in the metadata
        if parent_field is None:
            parent_field = self.parent_field

        self.store_subcluster_metadata(
            parent_field=parent_field, cluster_field=self._get_cluster_field_name()
        )

    def store_subcluster_metadata(self, parent_field: str, cluster_field: str):
        """
        Store subcluster metadata
        """
        return self.append_metadata_list(
            field="_subcluster_",
            value_to_append={
                "parent_field": parent_field,
                "cluster_field": cluster_field,
            },
            only_unique=True,
        )

    def subpartialfit_predict_update(
        self,
        dataset,
        vector_fields: list,
        filters: Optional[list] = None,
        cluster_ids: Optional[list] = None,
        verbose: bool = True,
    ):
        """

        Run partial fit subclustering on your dataset.

        Parameters
        ------------

        dataset: Dataset
            The dataset to call fit predict update on
        vector_fields: list
            The list of vector fields
        filters: list
            The list of filters

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()

            from relevanceai.package_utils.datasets import mock_documents
            ds = client.Dataset("sample")
            # Creates 100 sample documents
            documents = mock_documents(100)
            ds.upsert_documents(documents)

            from sklearn.cluster import MiniBatchKMeans
            model = MiniBatchKMeans(n_clusters=10)
            clusterer = ClusterOps(alias="minibatchkmeans-10", model=model)
            clusterer.subpartialfit_predict_update(
                dataset=ds,
            )

        """
        # Get data

        # self._init_dataset(dataset)

        filters = [] if filters is None else filters
        cluster_ids = [] if cluster_ids is None else cluster_ids
        # Loop through each unique cluster ID and run clustering
        parent_field = self.parent_field

        print("Getting unique cluster IDs...")
        unique_clusters = self.list_unique(parent_field)

        for i, unique_cluster in enumerate(tqdm(unique_clusters)):
            cluster_filters = filters.copy()
            cluster_filters += [
                {
                    "field": parent_field,
                    "filter_type": "category",
                    "condition": "==",
                    "condition_value": unique_cluster,
                }
            ]
            self.partial_fit_predict_update(
                dataset=self.dataset_id,
                vector_fields=vector_fields,
                filters=cluster_filters,
                verbose=False,
            )

        if verbose:
            print(
                "Build your clustering app here: "
                + f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
            )

    def list_unique(
        self,
        field: str = None,
        minimum_amount: int = 3,
        dataset_id: str = None,
        num_clusters: int = 1000,
    ):
        """
        List unique cluster IDS

        Example
        ---------

        .. code-block::

            from relevanceai import Client
            client = Client()
            cluster_ops = client.ClusterOps(
                alias="kmeans_8", vector_fields=["sample_vector_]
            )
            cluster_ops.list_unique()

        Parameters
        -------------
        alias: str
            The alias to use for clustering
        minimum_cluster_size: int
            The minimum size of the clusters
        dataset_id: str
            The dataset ID
        num_clusters: int
            The number of clusters

        """
        # Mainly to be used for subclustering
        # Get the cluster alias
        if dataset_id is None:
            self._check_for_dataset_id()
            dataset_id = self.dataset_id

        # currently the logic for facets is that when it runs out of pages
        # it just loops - therefore we need to store it in a simple hash
        # and then add them to a list
        all_cluster_ids: Set = set()

        while len(all_cluster_ids) < num_clusters:
            facet_results = self.datasets.facets(
                dataset_id=dataset_id,
                fields=[field],
                page_size=100,
                page=1,
                asc=True,
            )
            if "results" in facet_results:
                facet_results = facet_results["results"]
            if field not in facet_results:
                raise ValueError("Not valid field. Check schema.")
            for facet in facet_results[field]:
                if facet["frequency"] > minimum_amount:
                    curr_len = len(all_cluster_ids)
                    all_cluster_ids.add(facet[field])
                    new_len = len(all_cluster_ids)
                    if new_len == curr_len:
                        return list(all_cluster_ids)

        return list(all_cluster_ids)

    def subcluster_predict_documents(
        self,
        vector_fields: Optional[List] = None,
        filters: Optional[List] = None,
        min_parent_cluster_size: Optional[int] = None,
        cluster_ids: Optional[List] = None,
        verbose: bool = True,
    ):
        """
        Subclustering using fit predict update. This will loop through all of the
        different clusters and then run subclustering on them. For this, you need to

        Example
        ---------

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("sample")

            # Creating 100 sample documents
            from relevanceai.package_utils.datasets import mock_documents
            documents = mock_documents(100)
            ds.upsert_documents(documents)

            # Run simple clustering first
            ds.auto_cluster("kmeans-3", vector_fields=["sample_1_vector_"])

            # Start KMeans
            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=20)

            # Run subclustering.
            cluster_ops = client.ClusterOps(
                alias="subclusteringkmeans",
                model=model,
                parent_alias="kmeans-3")


        """
        filters = [] if filters is None else filters
        cluster_ids = [] if cluster_ids is None else cluster_ids
        # Loop through each unique cluster ID and run clustering

        if verbose:
            print("Getting unique cluster IDs...")
        if not cluster_ids:
            unique_clusters = self.list_unique(self.parent_field)
        else:
            unique_clusters = cluster_ids

        self._list_of_cluster_ops = []

        for i, unique_cluster in enumerate(tqdm(unique_clusters)):
            cluster_filters = filters.copy()
            cluster_filters += [
                {
                    "field": self.parent_field,
                    "filter_type": "category",
                    "condition": "==",
                    "condition_value": unique_cluster,
                }
            ]
            # Create a ClusterOps object with the filter
            if vector_fields is None:
                vector_fields = self.vector_fields
            ops = _SubClusterOps(
                credentials=self.credentials,
                model=self.model,
                vector_fields=vector_fields,
                alias=self.alias,
                dataset=self.dataset_id,
                parent_field=self.parent_field,
                n_clusters=None,
                cluster_config=None,
                outlier_value=self.outlier_value,
                outlier_label=self.outlier_label,
            )
            ops.run(
                dataset_id=self.dataset_id,
                vector_fields=vector_fields,
                filters=cluster_filters,
                verbose=False,
                min_parent_cluster_size=min_parent_cluster_size,
            )
            self._list_of_cluster_ops.append(ops)

        if verbose:
            print(
                "Build your clustering app here: "
                + f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
            )



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Union, List, Dict, Any, Optional
from scipy.cluster import hierarchy
from sklearn.neighbors import NearestNeighbors
from sklearn.metrics import (
    davies_bouldin_score,
    calinski_harabasz_score,
    silhouette_samples,
)
from sklearn.metrics.pairwise import (
    pairwise_distances,
)


class ClusterEvaluator:
    """
    Evaluate your clusters

    Parameters
    -------------

    X: np.ndarray
        The original data
    cluster_labels: List[str]
        A list of cluster labels
    centroids: Union[list, np.ndarray, str]
        The centroid vectors. If supplied it will use that. Otherwise, it will try to infer them
        from the model or calculate it based off the labels. To calculate mediods, let centroids = "mediods"
    model
        The model used for clustering.
    num_clusters: Optional[int]
        The number of clusters. This is required if we can't actually tell how many clusters there are
    outlier_label: Optional[str, int]
        The label if it is an outlier
    metric
        The metric to use for calculating distance, "euclidean", "cosine", ...
    verbose
        Whether to print stuff

    """

    def __init__(
        self,
        X: Union[list, np.ndarray],
        cluster_labels: Union[List[Union[str, float]], np.ndarray],
        centroids=None,
        # centroids: Union[list, np.ndarray, str] = None,
        cluster_names: Union[list, dict] = None,
        feature_names: Union[list, dict] = None,
        model=None,
        outlier_label: Union[str, int] = -1,
        metric: str = "euclidean",
        verbose: bool = False,
    ):
        if isinstance(X, list):
            self.X = np.array(X)
        else:
            self.X = X
        if isinstance(cluster_labels, list):
            self.cluster_labels = np.array(cluster_labels)
        else:
            self.cluster_labels = cluster_labels
        self.cluster_names = cluster_names
        self.feature_names = feature_names
        self.num_clusters = len(set(cluster_labels))
        self.model = model
        if centroids == "mediods":
            centroids = self._calculate_medoids(self.X, self.cluster_labels)
        if not centroids and not model:
            centroids = self._calculate_centroids(self.X, self.cluster_labels)
        elif not centroids and model:
            try:
                centroids = self._get_centroids_from_model(model)
            except:
                centroids = self._calculate_centroids(self.X, self.cluster_labels)
        if isinstance(centroids, (list, np.ndarray)):
            centroids = {i:c for i, c in enumerate(centroids)}
        if not isinstance(centroids, (list, dict, np.ndarray)):
            raise TypeError("centroid_vectors should be of type List or Numpy array")

        self.centroids = centroids
        self.outlier_label = outlier_label
        self.metric = metric
        if isinstance(self.centroids, dict):
            self.distance_matrix = pairwise_distances(
                list(self.centroids.values()), metric=self.metric
            )
        else:
            self.distance_matrix = pairwise_distances(
                self.centroids, metric=self.metric
            )
        self.verbose = verbose

    def _get_cluster_name(self, cluster_index):
        return (
            self.cluster_names[cluster_index]
            if self.cluster_names
            else f"cluster-{str(cluster_index)}"
        )

    def _get_feature_name(self, feature_index):
        return (
            self.feature_names[feature_index]
            if self.feature_names
            else f"feature-{str(feature_index)}"
        )

    def _get_centroids_from_model(self, model, output_format="array"):
        if hasattr(model, "cluster_centers_"):
            return model.cluster_centers_
        elif hasattr(model, "get_centers"):
            return model.get_centers()
        else:
            raise Exception

    def _calculate_centroids(self, X, cluster_labels):
        """Calculate the centroids"""
        centroid_vectors = {}
        for label in np.unique(cluster_labels):
            centroid_vectors[label] = X[cluster_labels == label].mean(axis=0)
        return centroid_vectors

    def _calculate_medoids(self, X, cluster_labels):
        centroids = self._calculate_centroids(X, cluster_labels)
        medoids = {}
        nbrs = NearestNeighbors(n_neighbors=1, algorithm="ball_tree").fit(X)
        medoid_indexes = nbrs.kneighbors(
            np.array(list(centroids.values())), n_neighbors=1, return_distance=False
        )
        medoids = X[medoid_indexes]
        return medoids

    def hierarchy_linkage(self, hierarchy_method: str = "ward"):
        self.linkage = hierarchy.linkage(
            list(self.centroids.values()), hierarchy_method
        )
        return self.linkage

    def pyplot_dendrogram(
        self, hierarchy_method=None, color_threshold=1.25, orientation="left", ax=None
    ):
        fig, ax = plt.subplots()
        if hierarchy_method:
            self.linkage = self.hierarchy_linkage(hierarchy_method=hierarchy_method)
        elif not hasattr(self, "linkage"):
            self.linkage = self.hierarchy_linkage(ax=ax)
        self.dendrogram = hierarchy.dendrogram(
            self.linkage,
            labels=self.cluster_names,
            orientation=orientation,
            color_threshold=color_threshold,
            ax=ax,
        )
        return fig

    def plotly_dendrogram(
        self, hierarchy_method=None, color_threshold=1.25, orientation="left"
    ):
        if hierarchy_method:
            self.linkage = self.hierarchy_linkage(hierarchy_method=hierarchy_method)
        elif not hasattr(self, "linkage"):
            self.linkage = self.hierarchy_linkage()
        try:
            import plotly.figure_factory as ff

            fig = ff.create_dendrogram(
                np.array(list(self.centroids.values())),
                labels=self.cluster_names,
                orientation=orientation,
                color_threshold=color_threshold,
                linkagefun=lambda x: self.linkage,
            )
            return fig
        except ImportError:
            raise ImportError(
                "This requires plotly installed, install with `pip install -U plotly`"
            )
        except TypeError as e:
            raise TypeError(
                e
                + " This is a common error that can be fixed with `pip install pyyaml==5.4.1`"
            )

    def plot_dendrogram(
        self,
        plot_method=None,
        hierarchy_method=None,
        color_threshold: float = 1,
        orientation: str = "left",
    ):
        if plot_method == "pyplot":
            return (
                self.pyplot_dendrogram(
                    hierarchy_method=hierarchy_method,
                    color_threshold=color_threshold,
                    orientation=orientation,
                ),
                "pyplot",
            )
        elif plot_method == "plotly":
            return (
                self.plotly_dendrogram(
                    hierarchy_method=hierarchy_method,
                    color_threshold=color_threshold,
                    orientation=orientation,
                ),
                "plotly",
            )
        else:
            try:
                return (
                    self.plotly_dendrogram(
                        hierarchy_method=hierarchy_method,
                        color_threshold=color_threshold,
                        orientation=orientation,
                    ),
                    "plotly",
                )
            except:
                return (
                    self.pyplot_dendrogram(
                        hierarchy_method=hierarchy_method,
                        color_threshold=color_threshold,
                        orientation=orientation,
                    ),
                    "pyplot",
                )

    def plot_distance_matrix(self, metric="euclidean", decimals=4):
        try:
            import plotly.express as px

            if metric == "euclidean":
                distance_matrix = self.distance_matrix
            else:
                if isinstance(self.centroids, dict):
                    distance_matrix = pairwise_distances(
                        list(self.centroids.values()), metric=metric
                    )
                else:
                    distance_matrix = pairwise_distances(
                        self.centroids, metric=metric
                    )
            
            if self.cluster_names:
                return (
                    px.imshow(
                        pd.DataFrame(
                            np.round(distance_matrix, 4),
                            columns=self.cluster_names,
                            index=self.cluster_names,
                        ),
                        text_auto=True,
                    ),
                    "plotly",
                )
            else:
                return (
                    px.imshow(np.round(distance_matrix, 4), text_auto=True),
                    "plotly",
                )
        except ImportError:
            raise ImportError(
                "This requires plotly installed, install with `pip install -U plotly`"
            )
        except TypeError as e:
            raise TypeError(
                e
                + " This is a common error that can be fixed with `pip install pyyaml==5.4.1`"
            )

    def plot_boxplot(self, summary_stats, name=""):
        try:
            import plotly.graph_objects as go
            fig = go.Figure()
            fig.add_trace(
                go.Box(
                    y=[
                        summary_stats['min'], 
                        summary_stats['25%'], 
                        summary_stats['50%'], 
                        summary_stats['75%'], 
                        summary_stats['max']
                    ], 
                    name=name
                )
            )
            return fig, "plotly"

        except ImportError:
            raise ImportError(
                "This requires plotly installed, install with `pip install -U plotly`"
            )

    @staticmethod
    def summary_statistics(array: np.ndarray, axis=0, simple=False):
        """
        Basic summary statistics
        """
        if axis == 2:
            return {
                "sum": array.sum(),
                "mean": array.mean(),
                "std": array.std(),
                "variance": array.var(),
                "min": array.min(),
                "max": array.max(),
                "25%": np.percentile(array, 25),
                "50%": np.percentile(array, 50),
                "75%": np.percentile(array, 75),
            }
        else:
            return {
                "sum": array.sum(axis=axis),
                "mean": array.mean(axis=axis),
                "std": array.std(axis=axis),
                "variance": array.var(axis=axis),
                "min": array.min(axis=axis),
                "max": array.max(axis=axis),
                "25%": np.percentile(array, 25, axis=axis),
                "50%": np.percentile(array, 50, axis=axis),
                "75%": np.percentile(array, 75, axis=axis),
            }

    def silhouette_samples(self):
        if not hasattr(self, "X_silhouette_samples"):
            self.X_silhouette_samples = silhouette_samples(
                self.X, self.cluster_labels, metric=self.metric
            )
        return self.X_silhouette_samples

    def silhouette_score(self):
        if not hasattr(self, "X_silhouette_samples"):
            self.silhouette_samples()
        return self.X_silhouette_samples.mean()

    def calinski_harabasz_score(self):
        return calinski_harabasz_score(self.X, self.cluster_labels)

    def davies_bouldin_score(self):
        return davies_bouldin_score(self.X, self.cluster_labels)

    @staticmethod
    def dunn_index(min_distance_from_centroid, max_centroid_distance):
        return min_distance_from_centroid / max_centroid_distance

    def distance_from_centroid(self, cluster_data, centroid):
        distances_from_centroid = pairwise_distances([centroid], cluster_data)
        return self.summary_statistics(distances_from_centroid, axis=2)

    def distance_from_centroid_to_another(self, other_cluster_data, centroid):
        """Store the distances from a centroid to another."""
        distances_from_centroid_to_another = pairwise_distances(
            [centroid], other_cluster_data
        )
        return self.summary_statistics(distances_from_centroid_to_another, axis=2)

    def squared_error_samples(self):
        if not hasattr(self, "X_squared_error_samples"):
            self.X_squared_error_samples = np.square(
                np.subtract(
                    [self.centroids[c] for c in self.cluster_labels],
                    self.X,
                )
            )
        return self.X_squared_error_samples

    def squared_error_score(self):
        if not hasattr(self, "X_squared_error_samples"):
            self.squared_error_samples()
        return self.X_squared_error_samples.sum()

    def mean_squared_error_score(self):
        if not hasattr(self, "X_silhouette_samples"):
            self.squared_error_samples()
        return self.X_squared_error_samples.mean()

    def squared_error_features_from_samples(self, squared_error_samples):
        squared_errors_by_feature = []
        for f in range(len(squared_error_samples[0])):
            squared_errors_by_feature.append(
                {
                    "feature_index": f,
                    "feature_id": self._get_feature_name(f),
                    "squared_errors": squared_error_samples[:, f].mean(),
                }
            )
        return squared_errors_by_feature

    @staticmethod
    def z_score(value, mean, std):
        return (value - mean) / std

    def closest_clusters(self, cluster_index, n_clusters):
        return [
            self._get_cluster_name(c)
            for c in self.distance_matrix[cluster_index].argsort()[:n_clusters]
        ]

    def furthest_clusters(self, cluster_index, n_clusters):
        return [
            self._get_cluster_name(c)
            for c in self.distance_matrix[cluster_index].argsort()[-n_clusters:][::-1]
        ]

    def internal_overview_report(
        self,
        store_centroids: bool = True,
        store_distance_matrix: bool = True,
        save=True,
    ):
        report = {}
        report["silhouette_score_summary"] = self.summary_statistics(
            self.silhouette_samples()
        )
        report["squared_error_summary"] = self.summary_statistics(
            self.squared_error_samples(), axis=2
        )
        report["calinski_harabasz_score"] = self.calinski_harabasz_score()
        report["davies_bouldin_score"] = self.davies_bouldin_score()
        report["silhouette_score"] = self.silhouette_score()
        report["total_squared_error_score"] = self.squared_error_score()
        report["mean_squared_error_score"] = self.mean_squared_error_score()
        # if self.model:
        #     report["_model_params"] = self.model.__dict__
        report["metric"] = self.metric
        report["num_clusters"] = self.num_clusters
        report["total_frequency"] = len(self.X)
        if self.cluster_names:
            report["cluster_names"] = self.cluster_names
        if self.feature_names:
            report["feature_names"] = self.feature_names
        if store_centroids:
            report["centroids"] = self.centroids
        if store_distance_matrix:
            report["distance_matrix"] = self.distance_matrix
        if save:
            self.report = report
        return report

    def internal_report(
        self,
        top_n_clusters: int = 5,
        top_n_features: int = 5,
        store_squared_errors: bool = True,
        store_distances: bool = True,
        store_centroids: bool = False,
        store_distance_matrix: bool = False,
        save=True,
    ):
        report = {}
        report.update(
            self.internal_overview_report(
                store_centroids=store_centroids,
                store_distance_matrix=store_distance_matrix,
            )
        )
        report["each_cluster_frequency"] = []
        report["each_cluster"] = []
        labels, label_counts = np.unique(self.cluster_labels, return_counts=True)
        for i, cluster_label in enumerate(labels):
            if cluster_label == self.outlier_label:
                cluster_name = "outlier_cluster"
            else:
                cluster_name = self._get_cluster_name(i)
            cluster_bool = self.cluster_labels == cluster_label
            current_cluster_data = self.X[cluster_bool]
            other_cluster_data = self.X[~cluster_bool]
            centroid_vector = self.centroids[cluster_label]
            current_summary = self.summary_statistics(current_cluster_data, axis=2)
            silouhette_summary = self.summary_statistics(
                self.X_silhouette_samples[cluster_bool], axis=2
            )
            cluster_document = {
                "cluster_id": cluster_name,
                "cluster_index": cluster_label,
                "centroid_vector_": centroid_vector,
                "frequency": label_counts[i],
                "summary": current_summary,
                "silouhette_score": silouhette_summary["mean"],
                "silouhette_score_summary": silouhette_summary,
                "closest_clusters": self.closest_clusters(i, top_n_clusters),
                "furthest_clusters": self.furthest_clusters(i, top_n_clusters),
                "features": {},
            }
            # squared errors
            if store_squared_errors:
                cluster_document["squared_error_summary"] = self.summary_statistics(
                    self.X_squared_error_samples[cluster_bool], axis=2
                )
                squared_error_by_features = self.squared_error_features_from_samples(
                    self.X_squared_error_samples[cluster_bool]
                )
                cluster_document["features"]["lowest_squared_errors"] = sorted(
                    squared_error_by_features,
                    key=lambda x: x["squared_errors"],
                    reverse=False,
                )[:top_n_features]
                cluster_document["features"]["highest_squared_errors"] = sorted(
                    squared_error_by_features,
                    key=lambda x: x["squared_errors"],
                    reverse=True,
                )[:top_n_features]
            # distances
            if store_distances:
                cluster_document[
                    "distance_from_centroid_summary"
                ] = self.distance_from_centroid(current_cluster_data, centroid_vector)
                cluster_document[
                    "distance_from_centroid_to_another_cluster_summary"
                ] = self.distance_from_centroid_to_another(
                    other_cluster_data, centroid_vector
                )
            report["each_cluster"].append(cluster_document)

        if store_distances:
            report["dunn_index_score"] = self.dunn_index(
                min_distance_from_centroid=min(
                    c["distance_from_centroid_summary"]["min"]
                    for c in report["each_cluster"]
                ),
                max_centroid_distance=self.distance_matrix.max(),
            )
        if save:
            self.report = report
        return report



"""
Transformers Pipeline Implementation
"""
from relevanceai.operations_new.transform_base import TransformBase
from typing import Optional


class TransformersPipelineTransform(TransformBase):
    def __init__(
        self, text_fields: list, pipeline, output_fields: Optional[str] = None, **kwargs
    ):
        self.text_fields = text_fields
        self.pipeline = pipeline
        self.task: str = pipeline.task
        self._name: str = pipeline.tokenizer.name_or_path
        self.output_fields = [
            self._generate_output_field(text_field)
            if output_fields is None
            else output_fields
            for text_field in text_fields
        ]
        for k, v in kwargs.items():
            setattr(self, k, v)
        print(f"Output fields are {self.output_fields}")

    @property
    def name(self):
        if self._name is not None:
            return self._name
        else:
            return "transformers-pipeline"

    def transform(self, documents):
        for i, text_field in enumerate(self.text_fields):
            texts = self.get_field_across_documents(text_field, documents)
            values = self.pipeline(texts)
            self.set_field_across_documents(self.output_fields[i], values, documents)
        return documents

    def _generate_output_field(self, text_field):
        return f"_{self.task}_.{self.name}." + text_field



"""
Sequential Workflows allow users to quickly swap in and out
new operations to test new things all while having different
operations logged.

Example
---------



"""
from typing import Union, Optional

from relevanceai.utils import DocUtils
from relevanceai.utils.logger import FileLogger
from relevanceai.dataset.dataset import Dataset


class Ops(DocUtils):
    """Base class for operations"""

    # @abstractmethod
    # def __call__(self, data):
    #     # Passes through documents or a dataset object
    #     raise NotImplementedError


class UpdateOp(DocUtils):
    """Inherit this class if you want documents to update
    at the end of this call
    """

    _update_: bool = True


class Input(Ops):
    # How do I link the original inputs to the outputs
    def __init__(
        self, input_fields: list, chunksize: int = 20, filters: Optional[list] = None
    ):
        self.input_fields = input_fields
        self.chunksize = chunksize
        self.filters = [] if filters is None else filters

    def __call__(self, dataset):
        return self.__iter__(dataset)

    def __iter__(self, dataset: Dataset):
        for c in dataset.chunk_dataset(
            chunksize=self.chunksize,
            select_fields=self.input_fields,
            filters=[
                {
                    "field": input_field,
                    "filter_type": "exists",
                    "condition": ">=",
                    "condition_value": " ",
                }
                for input_field in self.input_fields
            ]
            + self.filters,
        ):
            yield c
        return None


class Output(Ops):
    """Update documents"""

    _update_: bool = True

    def __init__(self, output_field: Union[str, list]):
        self.output_field = output_field

    def __call__(self, values, documents, dataset: Dataset):
        if isinstance(self.output_field, str):
            self.set_field_across_documents(self.output_field, values, documents)
        elif isinstance(self.output_field, list):
            for f in self.output_field:
                self.set_field_across_documents(f, values, documents)
        else:
            raise ValueError("Incorrect number of fields.")
        upsert_results = dataset.upsert_documents(documents)
        return upsert_results, documents


class SequentialWorkflow(DocUtils):
    """
    Defining a workflow with input and output fields and
    linking them up accordingly.

    Example
    ----------

    .. code-block::

        import random
        from relevanceai import Client
        from relevanceai.workflow.sequential import SequentialWorkflow

        client = Client()

        def vectorize(docs, *args, **kw):
            return [[random.randint(0, 100) for _ in range(3)] for _ in range(len(docs))]

        workflow = SequentialWorkflow(
            [
                Input(["sample_1_label"], chunksize=50),
                vectorize,
                Output("simple")
            ],
            log_filename="logs"
        )

        ds = client.Dataset("_mock_dataset_")
        workflow.run(ds, verbose=True)

    """

    def __init__(
        self, list_of_operations: list, log_filename: str = "workflow-run.txt"
    ):
        # TODO: make log filename generated otf
        self.list_of_operations = list_of_operations
        self.log_filename = log_filename

    def __call__(self, dataset):
        if isinstance(dataset, Dataset):
            # run on dataset
            return self.run(dataset)
        elif isinstance(dataset, list):
            return self.run_documents(dataset)

        raise ValueError(
            "Data type not supported. Please ensure it is a list of dicts of a Dataset object."
        )

    def run(self, dataset: Dataset, verbose: bool = False, log_to_file: bool = False):
        """
        Run the sequential workflow
        """
        if len(self.list_of_operations) == 0:
            return

        with FileLogger(self.log_filename, verbose=verbose, log_to_file=log_to_file):
            input_operator: Input = self.list_of_operations[0]

            if not hasattr(input_operator, "input_fields"):
                raise ValueError("The first operation needs to be Input.")

            # TODO: add progress bar
            for i, documents in enumerate(input_operator(dataset)):
                if documents is None:
                    return
                if len(input_operator.input_fields) == 1:
                    values = self.get_field_across_documents(
                        input_operator.input_fields[0], documents
                    )
                elif len(input_operator.input_fields) > 1:
                    values = self.get_fields_across_documents(
                        input_operator.input_fields[0], documents
                    )
                else:
                    raise ValueError("Input Operator")
                if verbose:
                    print(values)
                for j, op in enumerate(self.list_of_operations[1:]):
                    if hasattr(op, "run"):
                        values = op.run(values)
                        if verbose:
                            print(values)
                    elif hasattr(op, "_update_") and op._update_:
                        if verbose:
                            print("updating")
                            print(values)
                            print(documents)
                        self.set_field_across_documents(
                            op.output_field, values, documents
                        )
                        if verbose:
                            print(documents)
                        upsert_results = dataset.upsert_documents(documents)
                        if verbose:
                            print(upsert_results)
                    else:
                        values = op(values)
                        if verbose:
                            print(values)
        return



import warnings
from typing import Dict, List, Optional, Union


class ExplorerBase:
    def __init__(
        self,
        name: str,
        dataset,
        deployable_id: str = None,
        default_view: str = "charts",
        charts_view_column: int = 2,
        preview_centroids_page_size: int = 2,
        **kwargs,
    ):
        self.name = name
        # if not isinstance(dataset, Dataset):
        #     raise TypeError("'dataset' only accepts Relevance Dataset class. to create one use ds = client.Dataset('example').")
        self.dataset = dataset
        self.dataset_id = dataset.dataset_id
        self.deployable_id = deployable_id
        if default_view not in ["charts", "documents", "categories"]:
            raise KeyError(
                "'default_view' can only be one of ['charts', 'documents', 'categories']"
            )
        self.default_view = default_view
        app_config = None
        self.reloaded = False
        if deployable_id:
            try:
                app_config = self.dataset.get_app(deployable_id)
                self.reloaded = True
            except:
                raise Exception(
                    f"{deployable_id} does not exist in the dataset, the given id will be used for creating a new app."
                )
        if app_config:
            self.config = app_config["configuration"]
        else:
            self.config = {
                "dataset_name": self.dataset_id,
                "deployable_name": name,
                "type": "explore",
                "charts-view-columns": charts_view_column,
                "results-mode": default_view,
                "preview-centroids-page-size": preview_centroids_page_size,
                **kwargs,
            }

    def deploy(self, overwrite: bool = False):
        if self.deployable_id and self.reloaded:
            status = self.dataset.update_app(
                self.deployable_id, self.config, overwrite=overwrite
            )
            if status["status"] == "success":
                return self.dataset.get_app(self.deployable_id)
            else:
                raise Exception("Failed to update app")
        return self.dataset.create_app(self.config)

    def _append_or_replace(
        self, key: str, value, append: bool = True, default_value=None
    ):
        default_value = [] if default_value is None else default_value

        if key not in self.config:
            self.config[key] = default_value
        if append:
            self.config[key] += value
        else:
            self.config[key] = value

    def _agg_to_chart_config(
        self,
        groupby: List[Dict] = None,
        metrics: List[Dict] = None,
        sort: List[Dict] = None,
        page_size=None,
        chart_name: str = None,
        chart_mode: str = "column",
        show_frequency: bool = False,
        **kwargs,
    ):
        groupby = [] if groupby is None else groupby
        metrics = [] if metrics is None else metrics

        chart_config = {
            "groupby": groupby,
            "metrics": metrics,
            "chart-mode": chart_mode,
            "chart-name": chart_name,
            "show-frequency": show_frequency,
            **kwargs,
        }
        if not chart_name:
            chart_name = "Chart "  # can give more flexibility to this
            if groupby:
                groupby_names = [g["name"] for g in groupby]
                chart_name += f"grouped by ({', '.join(groupby_names)}) "
            if metrics:
                metrics_names = [m["name"] for m in metrics]
                chart_name += f"with metrics ({', '.join(metrics_names)}) "
            chart_config["chart-name"] = chart_name
        if sort:
            chart_config["metric-to-sort-by"] = self.dataset._return_sort_in_metrics(
                sort, metrics
            )
        if page_size:
            chart_config["page-size"] = page_size
        return chart_config



import warnings
from typing import List, Dict, Optional, Any, Union, Callable
from tqdm.auto import tqdm

from relevanceai.client.helpers import Credentials

from relevanceai.dataset.write import Write
from relevanceai.utils.decorators.analytics import track
from relevanceai.operations.vector.vectorizer import Vectorizer
from relevanceai.utils.logger import FileLogger

from relevanceai.dataset.io import IO


class Operations(Write, IO):
    def __init__(self, credentials: Credentials, dataset_id: str, **kwargs):
        self.credentials = credentials
        self.dataset_id = dataset_id
        super().__init__(credentials=self.credentials, dataset_id=dataset_id, **kwargs)

    @track
    def cluster(
        self,
        model: Any = None,
        vector_fields: Optional[List[str]] = None,
        alias: Optional[str] = None,
        filters: Optional[list] = None,
        include_cluster_report: bool = True,
        **kwargs,
    ):
        """
        Run clustering on your dataset.

        Example
        ----------

        .. code-block::

            from sklearn.cluster import KMeans
            model = KMeans()

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("sample")
            cluster_ops = ds.cluster(
                model=model, vector_fields=["sample_vector_"],
                alias="kmeans-8"
            )

        Parameters
        ------------

        model: Union[str, Any]
            Any K-Means model
        vector_fields: List[str]
            A list of possible vector fields
        alias: str
            The alias to be used to store your model
        cluster_config: dict
            The cluster config to use
            You can change the number of clusters for kmeans using:
            `cluster_config={"n_clusters": 10}`. For a full list of
            possible parameters for different models, simply check how
            the cluster models are instantiated.
        """
        from relevanceai.operations.cluster import ClusterOps

        ops = ClusterOps(
            credentials=self.credentials,
            model=model,
            alias=alias,
            vector_fields=vector_fields,
            verbose=False,
            **kwargs,
        )
        ops(
            dataset_id=self.dataset_id,
            vector_fields=vector_fields,
            include_cluster_report=include_cluster_report,
            filters=filters,
        )
        if alias is None:
            alias = ops.alias
        print(
            f"You can now utilise the ClusterOps object using `cluster_ops = client.ClusterOps(alias='{alias}', vector_fields={vector_fields}, dataset_id='{self.dataset_id}')`"
        )
        return ops

    @track
    def reduce_dims(
        self,
        alias: str,
        vector_fields: List[str],
        model: Any = "pca",
        n_components: int = 3,
        filters: Optional[list] = None,
        **kwargs,
    ):
        """
        Reduce dimensions

        Parameters
        --------------

        model: Callable
            model to reduce dimensions
        n_components: int
            The number of components
        alias: str
            The alias of the model
        vector_fields: List[str]
            The list of vector fields to support

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("sample")
            ds.reduce_dims(
                alias="sample",
                vector_fields=["sample_1_vector_"],
                model="pca"
            )

        """
        from relevanceai.operations.dr import ReduceDimensionsOps

        ops = ReduceDimensionsOps(
            credentials=self.credentials,
            model=model,
            n_components=n_components,
            **kwargs,
        )
        return ops.run(
            dataset_id=self.dataset_id,
            vector_fields=vector_fields,
            alias=alias,
            filters=filters,
        )

    dimensionality_reduction = reduce_dims

    @track
    def vectorize(
        self,
        fields: List[str] = None,
        filters: Optional[List] = None,
        **kwargs,
    ):
        """
        Vectorize the model

        Parameters
        ----------
        fields: List[str]
            A list of fields to vectorize

        encoders : Dict[str, List[Any]]
            A dictionary that creates a mapping between your unstructured fields
            and a list of encoders to run over those unstructured fields

        Returns
        -------
        dict
            If the vectorization process is successful, this dict contains
            the added vector names. Else, the dict is the request result
            containing error information.

        Example
        -------
        .. code-block::

            from relevanceai import Client

            client = Client()

            dataset_id = "sample_dataset_id"
            ds = client.Dataset(dataset_id)

            ds.vectorize(
                fields=["text_field_1", "text_field_2"],
                encoders={
                    "text": ["mpnet", "use"]
                }
            )

            # This operation with create 4 new vector fields
            #
            # text_field_1_mpnet_vector_, text_field_1_mpnet_vector_
            # text_field_1_use_vector_, text_field_1_use_vector_

        """
        if filters is None:
            filters = []
        from relevanceai.operations.vector import VectorizeOps

        ops = VectorizeOps(
            credentials=self.credentials,
            **kwargs,
        )

        return ops(
            dataset_id=self.dataset_id,
            fields=[] if fields is None else fields,
            filter=filters,
        )

    def advanced_vectorize(self, vectorizers: List[Vectorizer]):
        """
        Advanced vectorization.
        By setting an

        Example
        ----------

        .. code-block::

            # When first vectorizing
            from relevanceai.operations import Vectorizer
            vectorizer = Vectorizer(field="field_1", model=model, alias="value")
            ds.advanced_vectorize(
                [vectorizer],
            )

        Parameters
        -------------

        vectorize_mapping: dict
            Vectorize mapping

        """
        # TODO: Write test for advanced vectorize
        all_fields = [v.field for v in vectorizers]
        for vectorizer in tqdm(vectorizers):

            def encode(docs):
                docs = vectorizer.encode_documents(
                    fields=[vectorizer.field], documents=docs
                )
                return docs

            self.bulk_apply(
                dataset_id=self.dataset_id,
                bulk_func=encode,
                select_fields=all_fields,
            )

    @track
    def vector_search(self, **kwargs):
        """
        Allows you to leverage vector similarity search to create a semantic search engine. Powerful features of Relevance vector search:

        1. Multivector search that allows you to search with multiple vectors and give each vector a different weight.
        e.g. Search with a product image vector and text description vector to find the most similar products by what it looks like and what its described to do.
        You can also give weightings of each vector field towards the search, e.g. image_vector_ weights 100%, whilst description_vector_ 50% \n
            An example of a simple multivector query:

            >>> [
            >>>     {"vector": [0.12, 0.23, 0.34], "fields": ["name_vector_"], "alias":"text"},
            >>>     {"vector": [0.45, 0.56, 0.67], "fields": ["image_vector_"], "alias":"image"},
            >>> ]

            An example of a weighted multivector query:

            >>> [
            >>>     {"vector": [0.12, 0.23, 0.34], "fields": {"name_vector_":0.6}, "alias":"text"},
            >>>     {"vector": [0.45, 0.56, 0.67], "fields": {"image_vector_"0.4}, "alias":"image"},
            >>> ]

            An example of a weighted multivector query with multiple fields for each vector:

            >>> [
            >>>     {"vector": [0.12, 0.23, 0.34], "fields": {"name_vector_":0.6, "description_vector_":0.3}, "alias":"text"},
            >>>     {"vector": [0.45, 0.56, 0.67], "fields": {"image_vector_"0.4}, "alias":"image"},
            >>> ]

        2. Utilise faceted search with vector search. For information on how to apply facets/filters check out datasets.documents.get_where \n
        3. Sum Fields option to adjust whether you want multiple vectors to be combined in the scoring or compared in the scoring. e.g. image_vector_ + text_vector_ or image_vector_ vs text_vector_. \n
            When sum_fields=True:

            - Multi-vector search allows you to obtain search scores by taking the sum of these scores.
            - TextSearchScore + ImageSearchScore = SearchScore
            - We then rank by the new SearchScore, so for searching 1000 documents there will be 1000 search scores and results

            When sum_fields=False:

            - Multi vector search but not summing the score, instead including it in the comparison!
            - TextSearchScore = SearchScore1
            - ImagSearchScore = SearchScore2
            - We then rank by the 2 new SearchScore, so for searching 1000 documents there should be 2000 search scores and results.

        4. Personalization with positive and negative document ids.

            - For more information about the positive and negative document ids to personalize check out services.recommend.vector

        For more even more advanced configuration and customisation of vector search, reach out to us at dev@relevance.ai and learn about our new advanced_vector_search.

        Parameters
        ----------

        multivector_query : list
            Query for advance search that allows for multiple vector and field querying.
        positive_document_ids : dict
            Positive document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        negative_document_ids: dict
            Negative document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        vector_operation: string
            Aggregation for the vectors when using positive and negative document IDs, choose from ['mean', 'sum', 'min', 'max', 'divide', 'mulitple']
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        facets: list
            Fields to include in the facets, if [] then all
        filters: list
            Query for filtering the search results
        min_score: int
            Minimum score for similarity metric
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.
        include_vector: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        asc: bool
            Whether to sort results by ascending or descending order
        keep_search_history: bool
            Whether to store the history into Relevance. This will increase the storage costs over time.
        hundred_scale: bool
            Whether to scale up the metric by 100
        search_history_id: string
            Search history ID, only used for storing search histories.
        query: string
            What to store as the query name in the dashboard

        Example
        -----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("sample")
            results = ds.vector_search(multivector_query=MULTIVECTOR_QUERY)

        """

        from relevanceai.operations.vector import SearchOps

        ops = SearchOps(
            credentials=self.credentials,
            dataset_id=self.dataset_id,
        )

        return ops.vector_search(**kwargs)

    @track
    def hybrid_search(self, **kwargs):
        """
        Combine the best of both traditional keyword faceted search with semantic vector search to create the best search possible. \n

        For information on how to use vector search check out services.search.vector. \n

        For information on how to use traditional keyword faceted search check out services.search.traditional.

        Parameters
        ----------
        multivector_query : list
            Query for advance search that allows for multiple vector and field querying.
        text : string
            Text Search Query (not encoded as vector)
        fields : list
            Text fields to search against
        positive_document_ids : dict
            Positive document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        negative_document_ids: dict
            Negative document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        vector_operation: string
            Aggregation for the vectors when using positive and negative document IDs, choose from ['mean', 'sum', 'min', 'max', 'divide', 'mulitple']
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        facets: list
            Fields to include in the facets, if [] then all
        filters: list
            Query for filtering the search results
        min_score: float
            Minimum score for similarity metric
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.
        include_vector: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        asc: bool
            Whether to sort results by ascending or descending order
        keep_search_history: bool
            Whether to store the history into Relevance. This will increase the storage costs over time.
        hundred_scale: bool
            Whether to scale up the metric by 100
        search_history_id: string
            Search history ID, only used for storing search histories.
        edit_distance: int
            This refers to the amount of letters it takes to reach from 1 string to another string. e.g. band vs bant is a 1 word edit distance. Use -1 if you would like this to be automated.
        ignore_spaces: bool
            Whether to consider cases when there is a space in the word. E.g. Go Pro vs GoPro.
        traditional_weight: int
            Multiplier of traditional search score. A value of 0.025~0.075 is the ideal range


        Example
        -----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("sample")
            MULTIVECTOR_QUERY = [{"vector": [0, 1, 2], "fields": ["sample_vector_"]}]
            results = ds.vector_search(multivector_query=MULTIVECTOR_QUERY)

        """
        from relevanceai.operations.vector import SearchOps

        ops = SearchOps(
            credentials=self.credentials,
            dataset_id=self.dataset_id,
        )

        return ops.hybrid_search(**kwargs)

    def chunk_search(self, **kwargs):
        """
        Chunks are data that has been divided into different units. e.g. A paragraph is made of many sentence chunks, a sentence is made of many word chunks, an image frame in a video. By searching through chunks you can pinpoint more specifically where a match is occuring. When creating a chunk in your document use the suffix "chunk" and "chunkvector". An example of a document with chunks:

        >>> {
        >>>     "_id" : "123",
        >>>     "title" : "Lorem Ipsum Article",
        >>>     "description" : "Lorem Ipsum is simply dummy text of the printing and typesetting industry. Lorem Ipsum has been the industry's standard dummy text ever since the 1500s, when an unknown printer took a galley of type and scrambled it to make a type specimen book. It has survived not only five centuries, but also the leap into electronic typesetting, remaining essentially unchanged.",
        >>>     "description_vector_" : [1.1, 1.2, 1.3],
        >>>     "description_sentence_chunk_" : [
        >>>         {"sentence_id" : 0, "sentence_chunkvector_" : [0.1, 0.2, 0.3], "sentence" : "Lorem Ipsum is simply dummy text of the printing and typesetting industry."},
        >>>         {"sentence_id" : 1, "sentence_chunkvector_" : [0.4, 0.5, 0.6], "sentence" : "Lorem Ipsum has been the industry's standard dummy text ever since the 1500s, when an unknown printer took a galley of type and scrambled it to make a type specimen book."},
        >>>         {"sentence_id" : 2, "sentence_chunkvector_" : [0.7, 0.8, 0.9], "sentence" : "It has survived not only five centuries, but also the leap into electronic typesetting, remaining essentially unchanged."},
        >>>     ]
        >>> }

        For combining chunk search with other search check out services.search.advanced_chunk.

        Parameters
        ----------

        multivector_query : list
            Query for advance search that allows for multiple vector and field querying.
        chunk_field : string
            Field where the array of chunked documents are.
        chunk_scoring: string
            Scoring method for determining for ranking between document chunks.
        chunk_page_size: int
            Size of each page of chunk results
        chunk_page: int
            Page of the chunk results
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        facets: list
            Fields to include in the facets, if [] then all
        filters: list
            Query for filtering the search results
        min_score: int
            Minimum score for similarity metric
        include_vector: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        asc: bool
            Whether to sort results by ascending or descending order
        keep_search_history: bool
            Whether to store the history into Relevance. This will increase the storage costs over time.
        hundred_scale: bool
            Whether to scale up the metric by 100
        query: string
            What to store as the query name in the dashboard

        Example
        -----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("sample")
            results = ds.chunk_search(
                chunk_field="_chunk_",
                multivector_query=MULTIVECTOR_QUERY
            )

        """
        from relevanceai.operations.vector import SearchOps

        ops = SearchOps(
            credentials=self.credentials,
            dataset_id=self.dataset_id,
        )

        return ops.chunk_search(**kwargs)

    @track
    def multistep_chunk_search(self, **kwargs):
        """
        Multistep chunk search involves a vector search followed by chunk search, used to accelerate chunk searches or to identify context before delving into relevant chunks. e.g. Search against the paragraph vector first then sentence chunkvector after. \n

        For more information about chunk search check out datasets.search.chunk. \n

        For more information about vector search check out services.search.vector

        Example
        -----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("sample")
            results = ds.search.multistep_chunk(
                chunk_field="_chunk_",
                multivector_query=MULTIVECTOR_QUERY,
                first_step_multivector_query=FIRST_STEP_MULTIVECTOR_QUERY
            )

        Parameters
        ------------

        multivector_query : list
            Query for advance search that allows for multiple vector and field querying.
        chunk_field : string
            Field where the array of chunked documents are.
        chunk_scoring: string
            Scoring method for determining for ranking between document chunks.
        chunk_page_size: int
            Size of each page of chunk results
        chunk_page: int
            Page of the chunk results
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        facets: list
            Fields to include in the facets, if [] then all
        filters: list
            Query for filtering the search results
        min_score: int
            Minimum score for similarity metric
        include_vector: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        asc: bool
            Whether to sort results by ascending or descending order
        keep_search_history: bool
            Whether to store the history into Relevance. This will increase the storage costs over time.
        hundred_scale: bool
            Whether to scale up the metric by 100
        first_step_multivector_query: list
            Query for advance search that allows for multiple vector and field querying.
        first_step_page: int
            Page of the results
        first_step_page_size: int
            Size of each page of results
        query: string
            What to store as the query name in the dashboard

        """
        from relevanceai.operations.vector import SearchOps

        ops = SearchOps(
            credentials=self.credentials,
            dataset_id=self.dataset_id,
        )
        return ops.multistep_chunk_search(**kwargs)

    def launch_cluster_app(self, configuration: dict = None):
        """
        Launch an app with a given configuration


        Example
        --------

        .. code-block::

            ds.launch_cluster_app()

        Parameters
        -----------

        configuration: dict
            The configuration can be found in the deployable once created.

        """
        if configuration is None:
            url = f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
            print(
                "Build your clustering app here: "
                f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
            )
            return
        if "configuration" in configuration:
            configuration = configuration["configuration"]
        results = self.deployables.create(
            dataset_id=self.dataset_id, configuration=configuration
        )

        # After you have created an app
        url = f"https://cloud.relevance.ai/dataset/{results['dataset_id']}/deploy/cluster/{self.project}/{self.api_key}/{results['deployable_id']}/{self.region}"
        print(f"You can now access your deployable at {url}")
        return url

    @track
    def subcluster(
        self,
        model,
        alias: str,
        vector_fields,
        parent_field,
        filters: Optional[list] = None,
        cluster_ids: Optional[list] = None,
        min_parent_cluster_size: Optional[int] = None,
        **kwargs,
    ):
        from relevanceai.operations.cluster import SubClusterOps

        ops = SubClusterOps(
            model=model,
            credentials=self.credentials,
            alias=alias,
            vector_fields=vector_fields,
            dataset_id=self.dataset_id,
            parent_field=parent_field,
            dataset=self.dataset_id,
            **kwargs,
        )
        return ops.fit_predict(
            dataset=self.dataset_id,
            vector_fields=vector_fields,
            filters=filters,
            min_parent_cluster_size=min_parent_cluster_size,
            cluster_ids=cluster_ids,
        )

    @track
    def analyze_sentiment(
        self,
        text_fields: list,
        model_name: str = "siebert/sentiment-roberta-large-english",
        output_field: str = None,
        highlight: bool = False,
        positive_sentiment_name: str = "positive",
        max_number_of_shap_documents: Optional[int] = None,
        min_abs_score: float = 0.1,
        **apply_args,
    ):
        """
        Easily add sentiment to your dataset

        Example
        ----------

        .. code-block::

            ds.analyze_sentiment(field="sample_1_label")

        Parameters
        --------------

        field: str
            The field to add sentiment to
        output_field: str
            Where to store the sentiment values
        model_name: str
            The HuggingFace Model name.
        log_to_file: bool
            If True, puts the logs in a file. Otherwise, it will
        highlight: bool
            If True, this will include a SHAP explainer of what is causing positive
            and negative sentiment
        max_number_of_shap_documents: int
            The maximum number of shap documents
        min_abs_score: float
            The minimum absolute score for it to be considered important based on SHAP algorithm.

        """
        from relevanceai.operations_new.sentiment.ops import SentimentOps

        ops = SentimentOps(
            text_fields=text_fields,
            model_name=model_name,
            highlight=highlight,
            max_number_of_shap_documents=max_number_of_shap_documents,
            min_abs_score=min_abs_score,
        )

        return ops.run(self, batched=True)

        # return .fit_dataset(
        #     dataset=self,
        #     input_field=field,
        #     output_field=output_field,
        #     log_to_file=log_to_file,
        #     chunksize=chunksize,
        #     workflow_alias=workflow_alias,
        #     notes=notes,
        #     refresh=refresh,
        #     highlight=highlight,
        #     positive_sentiment_name=positive_sentiment_name,
        #     max_number_of_shap_documents=max_number_of_shap_documents,
        #     min_abs_score=min_abs_score,
        # )

    @track
    def question_answer(
        self,
        input_field: str,
        questions: Union[List[str], str],
        output_field: Optional[str] = None,
        model_name: str = "mrm8488/deberta-v3-base-finetuned-squadv2",
        verbose: bool = True,
        log_to_file: bool = True,
        filters: Optional[list] = None,
    ):
        """
        Question your dataset and retrieve answers from it.

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("ecommerce")
            ds.question_answer(
                input_field="product_title",
                question="What brand shoes",
                output_field="_question_test"
            )

        Parameters
        --------------

        field: str
            The field to add sentiment to
        output_field: str
            Where to store the sentiment values
        model_name: str
            The HuggingFace Model name.
        verbose: bool
            If True, prints progress bar workflow
        log_to_file: bool
            If True, puts the logs in a file.

        """
        from relevanceai.workflow.sequential import SequentialWorkflow, Input, Output
        from relevanceai.operations.text.qa.qa import QAOps

        if isinstance(questions, str):
            # Force listing so it loops through multiple question
            questions = [questions]

        model = QAOps(model_name=model_name)

        for question in tqdm(questions):
            print(f"Processing `{question}`...")

            def bulk_question_answer(contexts: list):
                return model.bulk_question_answer(question=question, contexts=contexts)

            if output_field is None:
                output_field = "_question_." + "-".join(
                    question.lower().strip().split()
                )
                print(f"No output field is detected. Setting to {output_field}")

            workflow = SequentialWorkflow(
                list_of_operations=[
                    Input([input_field], filters=filters),
                    bulk_question_answer,
                    Output(output_field),
                ]
            )
            workflow.run(self, verbose=verbose, log_to_file=log_to_file)

    # def summarize(
    #     self,
    #     summarize_fields: List[str],
    #     model_name: str = "sshleifer/distilbart-cnn-6-6",
    #     verbose: bool = True,
    #     log_to_file: bool = True,
    # ):
    #     """
    #     Question your dataset and retrieve answers from it.

    #     Example
    #     ----------

    #     .. code-block::

    #         from relevanceai import Client
    #         client = Client()
    #         ds = client.Dataset("ecommerce")
    #         ds.question_answer(
    #             input_field="product_title",
    #             question="What brand shoes",
    #             output_field="_question_test"
    #         )

    #     Parameters
    #     --------------

    #     field: str
    #         The field to add sentiment to
    #     output_field: str
    #         Where to store the sentiment values
    #     model_name: str
    #         The HuggingFace Model name.
    #     verbose: bool
    #         If True, prints progress bar workflow
    #     log_to_file: bool
    #         If True, puts the logs in a file.

    #     """
    #     from relevanceai.workflow.sequential import SequentialWorkflow, Input, Output
    #     from relevanceai.operations.text.qa.qa import QAOps

    #     model = QAOps(model_name=model_name)

    #     def bulk_question_answer(contexts: list):
    #         return model.bulk_question_answer(question=question, contexts=contexts)

    #     if output_field is None:
    #         output_field = "_question_." + "-".join(question.lower().strip().split())
    #         print(f"No output field is detected. Setting to {output_field}")

    #     workflow = SequentialWorkflow(
    #         list_of_operations=[
    #             Input([input_field]),
    #             bulk_question_answer,
    #             Output(output_field),
    #         ]
    #     )
    #     return workflow.run(self, verbose=verbose, log_to_file=log_to_file)

    def search(
        self,
        query: str = None,
        vector_search_query: Optional[dict] = None,
        fields_to_search: Optional[List] = None,
        select_fields: Optional[List] = None,
        include_vectors: bool = True,
        filters: Optional[List] = None,
        page: int = 0,
        page_size: int = 10,
        sort: dict = None,
        minimum_relevance: int = 0,
        query_config: dict = None,
        **kwargs,
    ):
        """
        Advanced Search

        Parameters
        -----------
        query: str
            The query to use
        vector_search_query: dict
            The vector search query
        fields_to_search: list
            The list of fields to search
        select_fields: list
            The fields to select

        """
        return self.datasets.fast_search(
            dataset_id=self.dataset_id,
            query=query,
            vectorSearchQuery=vector_search_query,
            fieldsToSearch=fields_to_search,
            includeFields=select_fields,
            includeVectors=include_vectors,
            filters=filters,
            page=page,
            pageSize=page_size,
            sort=sort,
            minimumRelevance=minimum_relevance,
            queryConfig=query_config,
            **kwargs,
        )

    advanced_search = search

    @track
    def list_deployables(self):
        """
        Use this function to list available deployables
        """
        return self.deployables.list()

    @track
    def train_text_model_with_gpl(
        self, text_field: str, title_field: Optional[str] = None
    ):
        """
        Train a text model using GPL (Generative Pseudo-Labelling)
        This can be helpful for `domain adaptation`.

        Example
        ---------

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("sample")
            ds.train_text_model(method="gpl")

        Parameters
        ------------

        text_field: str
            Text field

        """
        # The model can also be trained using this method
        from relevanceai.operations.text_finetuning import GPLOps

        ops = GPLOps.from_dataset(dataset=self)
        return ops.run(dataset=self, text_field=text_field, title_field=title_field)

    @track
    def train_text_model_with_tripleloss(
        self,
        text_field: str,
        label_field: str,
        output_dir: str = "trained_model",
        percentage_for_dev=None,
    ):
        """
        Supervised training a text model using tripleloss

        Example
        ---------

        .. code-block::
            from relevanceai import Client
            client = Client()
            ds = client.Dataset("ecommerce")
            ops = SupervisedTripleLossFinetuneOps.from_dataset(
                dataset=ds,
                base_model="distilbert-base-uncased",
                chunksize=16,
                triple_loss_type:str='BatchHardSoftMarginTripletLoss'
            )
            ops.run(text_field="detail_desc", label_field="_cluster_.desc_use_vector_.kmeans-10", output_dir)

        Parameters
        ------------

        text_field: str
            The field you want to use as input text for fine-tuning
        label_field: str
            The field indicating the classes of the input
        output_dir: str
            The path of the output directory
        percentage_for_dev: float
            a number between 0 and 1 showing how much of the data should be used for evaluation. No evaluation if None

        """
        # The model can also be trained using this method
        from relevanceai.operations.text_finetuning import (
            SupervisedTripleLossFinetuneOps,
        )

        ops = SupervisedTripleLossFinetuneOps.from_dataset(dataset=self)
        return ops.run(
            dataset=self,
            text_field=text_field,
            label_field=label_field,
            output_dir=output_dir,
            percentage_for_dev=percentage_for_dev,
        )

    def ClusterOps(self, alias, vector_fields: List, verbose: bool = False, **kwargs):
        """
        ClusterOps object
        """
        from relevanceai import ClusterOps

        return ClusterOps(
            credentials=self.credentials,
            alias=alias,
            vector_fields=vector_fields,
            dataset_id=self.dataset_id,
            verbose=verbose,
            **kwargs,
        )

    @track
    def label_from_list(
        self,
        vector_field: str,
        model: Callable,
        label_list: list,
        similarity_metric="cosine",
        number_of_labels: int = 1,
        score_field: str = "_search_score",
        alias: Optional[str] = None,
    ):
        """Label from a given list.

        Parameters
        ------------

        vector_field: str
            The vector field to label in the original dataset
        model: Callable
            This will take a list of strings and then encode them
        label_list: List
            A list of labels to accept
        similarity_metric: str
            The similarity metric to accept
        number_of_labels: int
            The number of labels to accept
        score_field: str
            What to call the scoring of the labels
        alias: str
            The alias of the labels

        Example
        --------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample")

            # Get a model to help us encode
            from vectorhub.encoders.text.tfhub import USE2Vec
            enc = USE2Vec()

            # Use that model to help with encoding
            label_list = ["dog", "cat"]

            df = client.Dataset("_github_repo_vectorai")

            df.label_from_list("documentation_vector_", enc.bulk_encode, label_list, alias="pets")

        """
        if alias is None:
            warnings.warn(
                "No alias is detected for labelling. Default to 'default' as the alias."
            )
            alias = "default"
        print("Encoding labels...")
        label_vectors = []
        for c in self.chunk(label_list, chunksize=20):
            with FileLogger(verbose=True):
                label_vectors.extend(model(c))

        if len(label_vectors) == 0:
            raise ValueError("Failed to encode.")

        # we need this to mock label documents - these values are not important
        # and can be changed :)
        LABEL_VECTOR_FIELD = "label_vector_"
        LABEL_FIELD = "label"

        label_documents = [
            {LABEL_VECTOR_FIELD: label_vectors[i], LABEL_FIELD: label}
            for i, label in enumerate(label_list)
        ]

        return self._bulk_label_dataset(
            label_documents=label_documents,
            vector_field=vector_field,
            label_vector_field=LABEL_VECTOR_FIELD,
            similarity_metric=similarity_metric,
            number_of_labels=number_of_labels,
            score_field=score_field,
            label_fields=[LABEL_FIELD],
            alias=alias,
        )

    def _bulk_label_dataset(
        self,
        label_documents,
        vector_field,
        label_vector_field,
        similarity_metric,
        number_of_labels,
        score_field,
        label_fields,
        alias,
    ):
        def label_and_store(d: dict):
            labels = self._get_nearest_labels(
                label_documents=label_documents,
                vector=self.get_field(vector_field, d),
                label_vector_field=label_vector_field,
                similarity_metric=similarity_metric,
                number_of_labels=number_of_labels,
                score_field=score_field,
                label_fields=label_fields,
            )
            d.update(self._store_labels_in_document(labels, alias))
            return d

        def bulk_label_documents(documents):
            [label_and_store(d) for d in documents]
            return documents

        print("Labelling dataset...")
        return self.bulk_apply(
            bulk_label_documents,
            filters=[
                {
                    "field": vector_field,
                    "filter_type": "exists",
                    "condition": ">=",
                    "condition_value": " ",
                },
            ],
            select_fields=[vector_field],
        )

    @track
    def _store_labels_in_document(self, labels: list, alias: str):
        if isinstance(labels, dict) and "label" in labels:
            return {"_label_": {alias: labels["label"]}}
        return {"_label_": {alias: labels}}

    def _get_nearest_labels(
        self,
        label_documents: List[Dict],
        vector: List[float],
        label_vector_field: str,
        similarity_metric,
        number_of_labels: int,
        label_fields: List[str],
        score_field="_label_score",
    ):

        from relevanceai.operations.vector.local_nearest_neighbours import (
            NearestNeighbours,
        )

        nearest_neighbors: List[Dict] = NearestNeighbours.get_nearest_neighbours(
            label_documents,
            vector,
            label_vector_field,
            similarity_metric,
            score_field=score_field,
        )[:number_of_labels]
        labels: List[Dict] = self.subset_documents(
            [score_field] + label_fields, nearest_neighbors
        )
        # Preprocess labels for easier frontend access
        new_labels = {}
        for lf in label_fields:
            new_labels[lf] = [
                {"label": l.get(lf), score_field: l.get(score_field)} for l in labels
            ]
        return new_labels



"""Simple test for testing cluster ops
"""
import pytest
from relevanceai.dataset import Dataset


@pytest.mark.skip(reason="not implemented properly")
def test_cluster_ops_explain_closest(test_dataset: Dataset):
    # Get the cluster ops object
    # Using the cluster ops object, highlight the closest ones
    # highlight the closest ones

    VECTOR_FIELDS = ["sample_1_vector_"]
    ALIAS = "explainer"

    cluster_ops = test_dataset.cluster(vector_fields=VECTOR_FIELDS, alias=ALIAS)

    from relevanceai.operations_new.cluster.ops import ClusterOps as NewClusterOps

    new_cluster_ops = NewClusterOps(
        dataset_id=test_dataset.dataset_id, vector_fields=VECTOR_FIELDS, alias=ALIAS
    )
    results = new_cluster_ops.explain_text_clusters(
        text_field="sample_1_label", n_closest=5
    )
    assert results



"""
Workflows Client
"""
from relevanceai.utils.base import _Base
from enum import Enum
from typing import Union


class WorkflowStatus(Enum):
    IN_PROGRESS: str = "inprogress"
    COMPLETED: str = "complete"
    FAILED: str = "failed"


class WorkflowsClient(_Base):
    def __init__(self, credentials):
        super().__init__(credentials)

    def trigger(self, params: dict, notebook_path: str, instance_type: str):
        """
        Trigger a workflow
        """
        return self.make_http_request(
            "/workflows/trigger",
            method="POST",
            parameters={
                "params": params,
                "notebook_path": notebook_path,
                "instance_type": instance_type,
            },
        )

    def list(self):
        return self.make_http_request(
            "/workflows/list",
            method="GET",
        )

    def get(self, workflow_id: str):
        """
        Returns

            {
                "job_status": "InProgress",
                "job_message": "string",
                "creation_time": "string",
                "notebook_path": "string",
                "params": { },
                "_id": "string",
                "metadata": { }
            }
        """
        return self.make_http_request(f"/workflows/{workflow_id}/get", method="POST")

    def metadata(self, workflow_id: str, metadata: dict):
        """
        Update metadata for a workflow run
        """
        return self.make_http_request(
            f"/workflows/{workflow_id}/metadata",
            method="POST",
            parameters={"metadata": metadata},
        )

    def status(
        self,
        workflow_id: str,
        metadata: dict,
        workflow_name: str,
        additional_information: str = "",
        status: Union[WorkflowStatus, str] = WorkflowStatus.IN_PROGRESS,
    ):
        """
        If status is complete, it triggers an email.
        """
        if isinstance(status, WorkflowStatus):
            status = status.value
        return self.make_http_request(
            f"/workflows/{workflow_id}/status",
            method="POST",
            parameters={
                "metadata": {},
                "status": status,
                "workflow_name": workflow_name,
                "additional_information": additional_information,
                "metadata": metadata,
            },
        )



from typing import List, Union

import numpy as np

from relevanceai.operations_new.dr.models.base import DimReductionModelBase

try:
    from ivis import Ivis
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(
        f"{e}\nInstall ivis\n \
        CPU: pip install -U ivis[cpu]\n \
        GPU: pip install -U ivis[gpu]"
    )


class IvisModel(DimReductionModelBase):
    def __init__(
        self,
        n_components: int,
        alias: Union[str, None],
        **kwargs,
    ):
        self.model = Ivis(embedding_dims=n_components, **kwargs)
        self.model_name = "ivis"
        self.alias = alias

    def fit(
        self,
        vectors: Union[List[List[float]], np.ndarray],
    ) -> None:
        """The function takes a list of lists of floats and fits the model to the data

        Parameters
        ----------
        vectors : Union[List[List[float]], np.ndarray]
            Union[List[List[float]], np.ndarray]

        """

        if isinstance(vectors, list):
            vectors = np.array(vectors)

        if self.model.chunksize > vectors.shape[0]:
            self.model.chunksize = vectors.shape[0]

        self.model.fit(vectors)

    def fit_transform(
        self,
        vectors: List[List[float]],
    ) -> List[List[float]]:
        """It takes a list of vectors, fits the model to the vectors, and then transforms the vectors

        Parameters
        ----------
        vectors : List[List[float]]
            List[List[float]]

        Returns
        -------
            A list of lists of floats.

        """

        vectors = np.array(vectors)
        self.model.fit(vectors)
        reduce_vectors = self.model.transform(vectors)
        return reduce_vectors.tolist()



from typing import Any, List, Union

from relevanceai.operations_new.vectorize.models.base import VectorizeModelBase
from relevanceai.utils.decorators.vectors import catch_errors


class SentenceTransformer2Vec(VectorizeModelBase):
    def __init__(
        self,
        model: Any,
        vector_length: Union[None, int],
        model_name: str,
    ):
        from sentence_transformers import SentenceTransformer

        self.model: SentenceTransformer = model
        if vector_length is None:
            self.vector_length = self._get_vector_length()
        else:
            self.vector_length = vector_length
        self.model_name = model_name

    def _get_vector_length(self) -> int:
        """It takes a string, encodes it into a vector, and returns the length of that vector

        Returns
        -------
            The length of the vector.

        """
        vector = self.model.encode(["test"]).tolist()[0]
        length = len(vector)
        return length

    @catch_errors
    def encode(self, text: str) -> List[float]:
        """The function takes a string as input and returns a list of floats

        Parameters
        ----------
        text : str
            The text to be encoded.

        Returns
        -------
            A list of floats.

        """
        return self.model.encode([text])[0].tolist()

    @catch_errors
    def bulk_encode(self, texts: List[str]) -> List[List[float]]:
        """The function takes a list of strings and returns a list of lists of floats

        Parameters
        ----------
        texts : List[str]
            List[str]

        Returns
        -------
            A list of lists of floats.

        """
        return self.model.encode(texts).tolist()



"""
Labelling performs a vector search on the labels and fetches the closest
max_number_of_labels.

"""
from copy import deepcopy
from tkinter import N
from typing import Any, Dict, List

from numpy import maximum

from relevanceai.operations_new.transform_base import TransformBase


class TextTagTransform(TransformBase):
    def __init__(
        self,
        text_field: str,
        labels: list,
        minimum_score: float = 0.25,
        model_id=None,
        maximum_number_of_labels: int = 5,
        **kwargs,
    ):
        self.text_field = text_field
        self.labels = labels
        self.model_id = model_id
        self.minimum_score = minimum_score
        self.maximum_number_of_labels = maximum_number_of_labels
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def classifier(self):
        if not hasattr(self, "_classifier"):
            from transformers import pipeline

            try:
                self._classifier = pipeline(
                    "zero-shot-classification", model=self.model_id, device=0
                )
            except:
                self._classifier = pipeline(
                    "zero-shot-classification", model=self.model_id, device=-1
                )
        return self._classifier

    def transform(  # type: ignore
        self,
        documents,
    ) -> List[Dict[str, Any]]:
        new_docs = []
        for doc in documents:
            query = self.get_field(self.text_field, doc)
            labels = self.tag_text(query, self.labels)
            new_doc = {"_id": doc["_id"]}
            self.set_field(
                self._generate_output_field(self.text_field), new_doc, labels
            )
            new_docs.append(new_doc)

        return new_docs

    @property
    def name(self):
        return "texttag"

    def tag_text(
        self,
        query,
        labels,
    ):
        """It takes a query vector, a vector field, a list of documents, and a few other parameters, and
        returns a list of documents sorted by their cosine similarity to the query vector

        Parameters
        ----------
        query_vector
            the vector you want to compare against
        vector_field
            the field in the documents that contains the vector
        documents
            list of documents
        reverse, optional
            True/False
        score_field : str, optional
            str = "_label_score"
        max_number_of_labels : int, optional
            int = 1,
        similarity_threshold : float, optional
            float = 0,

        Returns
        -------
            A list of dictionaries.

        """
        output = self.classifier(query, labels, multi_label=True)
        labels = [
            l
            for i, l in enumerate(output["labels"])
            if output["scores"][i] > self.minimum_score
        ]

        counter = 0
        new_labels = []
        for i, label in enumerate(labels):
            if (
                self.maximum_number_of_labels is not None
                and counter >= self.maximum_number_of_labels
            ):
                return new_labels

            if label not in new_labels:
                new_labels.append(label)
                counter += 1

        return new_labels

    def get_operation_metadata(self) -> Dict[str, Any]:
        return dict(
            operation="texttags",
            values=str(
                {
                    "minimum_score": self.minimum_score,
                    "max_number_of_labels": self.maximum_number_of_labels,
                }
            ),
        )

    def _generate_output_field(self, field):
        return f"_{self.name}_.{field.lower().replace(' ', '_')}.{self.model_id}"



import traceback
import warnings
import numpy as np
import pandas as pd
from copy import deepcopy

from typing import Optional, Union, Callable, Dict, Any, Set, List

from relevanceai.utils.decorators.analytics import track

from relevanceai.operations_new.ops_base import OperationAPIBase
from relevanceai.operations_new.cluster.transform import ClusterTransform

from relevanceai.constants import Warning
from relevanceai.constants.errors import MissingClusterError
from relevanceai.constants import MissingClusterError, Warning


class ClusterOps(ClusterTransform, OperationAPIBase):
    """
    Cluster-related functionalities
    """

    # These need to be instantiated on __init__
    model_name: str

    def __init__(
        self,
        dataset_id: str,
        vector_fields: list,
        alias: str,
        model=None,
        model_kwargs=None,
        cluster_field: str = "_cluster_",
        byo_cluster_field: str = None,
        include_cluster_report: bool = False,
        verbose: bool = False,
        **kwargs,
    ):
        """
        ClusterOps objects
        """
        self.dataset_id = dataset_id
        self.vector_fields = vector_fields
        self.cluster_field = cluster_field
        self.verbose = verbose
        self.model = model
        if isinstance(self.model, str):
            self.model_name = self.model
        else:
            self.model_name = str(self.model)

        if model_kwargs is None:
            model_kwargs = {}

        self.model_kwargs = model_kwargs

        self.include_cluster_report = include_cluster_report

        for k, v in kwargs.items():
            setattr(self, k, v)

        super().__init__(
            dataset_id=dataset_id,
            vector_fields=vector_fields,
            alias=alias,
            cluster_field=cluster_field,
            verbose=verbose,
            model=model,
            model_kwargs=model_kwargs,
            include_cluster_report=include_cluster_report,
            **kwargs,
        )

        # alias is set after model so that we can get the number of clusters
        # if the model needs to be instantiated
        self.alias = self._get_alias(alias)

        self.byo_cluster_field = byo_cluster_field
        if byo_cluster_field is not None:
            self.create_byo_clusters()

    def post_run(self, dataset, documents, updated_documents):
        centroid_documents = self.get_centroid_documents()
        self.insert_centroids(centroid_documents)
        if hasattr(self, "include_cluster_report") and self.include_cluster_report:
            try:
                from relevanceai.recipes.model_observability.cluster.report import (
                    ClusterReport,
                )

                app = ClusterReport(f"Cluster Report for {self.alias}", dataset)
                app.start_cluster_evaluator(
                    self.get_field_across_documents(self.vector_fields[0], documents),
                    self.get_field_across_documents(
                        self._get_cluster_field_name(), updated_documents
                    ),
                    # centroids=centroid_documents
                )
                app.evaluator.X_silhouette_samples = np.array(
                    self.get_field_across_documents(
                        self._silhouette_score_field_name(), updated_documents
                    )
                )
                app.evaluator.X_squared_error_samples = np.array(
                    self.get_field_across_documents(
                        self._squared_error_field_name(), updated_documents
                    )
                )
                app.section_cluster_report()
                print()
                print("We've built your cluster report app:")
                app.deploy()
            except Exception as e:
                print(e)
                print("Couldnt' create cluster report.")
        return

    def insert_centroids(
        self,
        centroid_documents,
    ) -> None:
        """
        Insert centroids
        Centroids look below

        .. code-block::

            cluster_ops = client.ClusterOps(
                vector_field=["sample_1_vector_"],
                alias="sample"
            )
            cluster_ops.insert_centroids(
                centorid_documents=[
                    {"_id" : "cluster-0", "sample_1_vector_": [1, 1, 1]},
                    {"_id" : "cluster-1", "sample_1_vector_": [1, 2, 2]},
                ]
            )

        """
        res = self.datasets.cluster.centroids.insert(
            dataset_id=self.dataset_id,
            cluster_centers=self.json_encoder(centroid_documents),
            vector_fields=self.vector_fields,
            alias=self.alias,
        )
        return res

    def calculate_centroids(self, method="mean"):
        """
        calculates the centroids from the dataset vectors
        """

        # calculate the centroids
        centroid_vectors = {}

        def calculate_centroid(vectors):
            X = np.array(vectors)
            return X.mean(axis=0)

        centroid_vectors = self._operate_across_clusters(
            field=self.vector_fields[0], func=calculate_centroid
        )

        if isinstance(centroid_vectors, dict):
            centroid_vectors = [
                {"_id": k, self.vector_fields[0]: v}
                for k, v in centroid_vectors.items()
            ]
        return centroid_vectors

    def create_centroids(self, insert: bool = True):
        """
        Calculate centroids from your dataset vectors.

        Example
        --------

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("sample")
            cluster_ops = ds.ClusterOps(
                alias="kmeans-25",
                vector_fields=['sample_vector_']
            )
            centroids = cluster_ops.create_centroids()

        """
        # Get an array of the different vectors
        if len(self.vector_fields) > 1:
            raise NotImplementedError(
                "Do not currently support multiple vector fields for centroid creation."
            )

        # calculate the centroids
        centroid_vectors = self.calculate_centroids()

        if insert:
            self.insert_centroids(
                centroid_documents=centroid_vectors,
            )
        return centroid_vectors

    def get_centroid_documents(self):
        centroid_vectors = {}
        if hasattr(self.model, "_centroids") and self.model._centroids is not None:
            # TODO: fix this so that it creates the proper labels
            centroid_vectors = self.model._centroids.tolist()
            # get the cluster label function
            labels = range(len(centroid_vectors))
            cluster_ids = self.format_cluster_labels(labels)
            if len(self.vector_fields) > 1:
                warnings.warn(
                    "Currently do not support inserting centroids with multiple vector fields"
                )
            centroids = [
                {"_id": k, self.vector_fields[0]: v}
                for k, v in zip(cluster_ids, centroid_vectors)
            ]
        else:
            centroids = self.create_centroids()
        return centroids

    @property
    def centroids(self):
        """
        Access the centroids of your dataset easily

        .. code-block::

            ds = client.Dataset("sample")
            cluster_ops = ds.ClusterOps(
                vector_fields=["sample_vector_"],
                alias="simple"
            )
            cluster_ops.centroids

        """
        if not hasattr(self, "_centroids"):
            self._centroids = self.datasets.cluster.centroids.documents(
                dataset_id=self.dataset_id,
                vector_fields=self.vector_fields,
                alias=self.alias,
                page_size=9999,
                include_vector=True,
            )["results"]
        return self._centroids

    def get_centroid_from_id(
        self,
        cluster_id: str,
    ) -> Dict[str, Any]:
        """> It takes a cluster id and returns the centroid with that id

        Parameters
        ----------
        cluster_id : str
            The id of the cluster to get the centroid for.

        Returns
        -------
            The centroid with the given id.

        """

        for centroid in self.centroids:
            if centroid["_id"] == cluster_id:
                return centroid

        raise ValueError(f"Missing the centroid with id {cluster_id}")

    def list_cluster_ids(
        self,
        alias: str = None,
        minimum_cluster_size: int = 0,
        num_clusters: int = 1000,
    ):
        """
        List unique cluster IDS

        Example
        ---------

        .. code-block::

            from relevanceai import Client
            client = Client()
            cluster_ops = client.ClusterOps(
                alias="kmeans_8", vector_fields=["sample_vector_]
            )
            cluster_ops.list_cluster_ids()

        Parameters
        -------------
        alias: str
            The alias to use for clustering
        minimum_cluster_size: int
            The minimum size of the clusters
        num_clusters: int
            The number of clusters

        """
        if alias is None:
            alias = self.alias
        # Mainly to be used for subclustering
        # Get the cluster alias
        cluster_field = self._get_cluster_field_name()

        # currently the logic for facets is that when it runs out of pages
        # it just loops - therefore we need to store it in a simple hash
        # and then add them to a list
        all_cluster_ids: Set = set()

        while len(all_cluster_ids) < num_clusters:
            facet_results = self.datasets.facets(
                dataset_id=self.dataset_id,
                fields=[cluster_field],
                page_size=int(self.config["data.max_clusters"]),
                page=1,
                asc=True,
            )
            if "results" in facet_results:
                facet_results = facet_results["results"]
            if cluster_field not in facet_results:
                raise MissingClusterError(alias=alias)
            for facet in facet_results[cluster_field]:
                if facet["frequency"] > minimum_cluster_size:
                    curr_len = len(all_cluster_ids)
                    all_cluster_ids.add(facet[cluster_field])
                    new_len = len(all_cluster_ids)
                    if new_len == curr_len:
                        return list(all_cluster_ids)

        return list(all_cluster_ids)

    def list_closest(
        self,
        cluster_ids: Optional[list] = None,
        select_fields: Optional[List] = None,
        approx: int = 0,
        page_size: int = 1,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: Optional[list] = None,
        facets: Optional[list] = None,
        include_vector: bool = False,
        cluster_properties_filters: Optional[Dict] = None,
        include_count: bool = False,
        include_facets: bool = False,
        verbose: bool = False,
    ):
        """
        List of documents closest from the center.
        Parameters
        ----------
        dataset_id: string
            Unique name of dataset
        vector_fields: list
            The vector fields where a clustering task runs
        cluster_ids: list
            Any of the cluster ids
        alias: string
            Alias is used to name a cluster
        centroid_vector_fields: list
            Vector fields stored
        select_fields: list
            Fields to include in the search results, empty array/list means all fields
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        facets: list
            Fields to include in the facets, if [] then all
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        include_facets: bool
            Include facets in the search results
        cluster_properties_filter: dict
            Filter if clusters with certain characteristics should be hidden in results
        """
        if cluster_properties_filters is None:
            cluster_properties_filters = {}
        return self.datasets.cluster.centroids.list_closest_to_center(
            dataset_id=self.dataset_id,
            vector_fields=self.vector_fields,
            alias=self.alias,
            cluster_ids=cluster_ids,
            select_fields=select_fields,
            approx=approx,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            filters=filters,
            facets=facets,
            include_vector=include_vector,
            include_count=include_count,
            include_facets=include_facets,
            cluster_properties_filter=cluster_properties_filters,
            verbose=verbose,
        )

    def list_furthest(
        self,
        cluster_ids: Optional[List] = None,
        centroid_vector_fields: Optional[List] = None,
        select_fields: Optional[List] = None,
        approx: int = 0,
        sum_fields: bool = True,
        page_size: int = 3,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: Optional[List] = None,
        # facets: List = [],
        min_score: int = 0,
        include_vector: bool = False,
        include_count: bool = True,
        cluster_properties_filter: Optional[Dict] = {},
    ):
        """
        List documents furthest from the center.

        Parameters
        ----------
        dataset_id: string
            Unique name of dataset
        vector_fields: list
            The vector field where a clustering task was run.
        cluster_ids: list
            Any of the cluster ids
        alias: string
            Alias is used to name a cluster
        select_fields: list
            Fields to include in the search results, empty array/list means all fields
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        facets: list
            Fields to include in the facets, if [] then all
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        include_facets: bool
            Include facets in the search results
        """
        return self.datasets.cluster.centroids.list_furthest_from_center(
            dataset_id=self.dataset_id,
            vector_fields=self.vector_fields,
            alias=self.alias,
            cluster_ids=cluster_ids,
            centroid_vector_fields=centroid_vector_fields,
            select_fields=select_fields,
            approx=approx,
            sum_fields=sum_fields,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            filters=filters,
            min_score=min_score,
            include_vector=include_vector,
            include_count=include_count,
            cluster_properties_filter=cluster_properties_filter,
        )

    def store_operation_metadatas(self):
        self.store_operation_metadata(
            operation="cluster",
            values=str(
                {
                    "model": self.model,
                    "vector_fields": self.vector_fields,
                    "alias": self.alias,
                    "model_kwargs": self.model_kwargs,
                }
            ),
        )

    @staticmethod
    def _get_filters(
        filters: List[Dict[str, Union[str, int]]],
        vector_fields: List[str],
    ) -> List[Dict[str, Union[str, int]]]:
        """It takes a list of filters and a list of vector fields and returns a list of filters that
        includes the original filters and a filter for each vector field that checks if the vector field
        exists

        Parameters
        ----------
        filters : List[Dict[str, Union[str, int]]]
            List[Dict[str, Union[str, int]]]
        vector_fields : List[str]
            List[str] = ["vector_field_1", "vector_field_2"]

        Returns
        -------
            A list of dictionaries.

        """

        vector_field_filters = [
            {
                "field": vector_field,
                "filter_type": "exists",
                "condition": ">=",
                "condition_value": " ",
            }
            for vector_field in vector_fields
        ]

        filters = deepcopy(filters)

        if filters is None:
            filters = vector_field_filters
        else:
            filters += vector_field_filters  # type: ignore

        return filters

    def merge(self, target_cluster_id: str, cluster_ids: list):
        """
        Merge clusters into the target cluster.
        The centroids are re-calculated and become a new middle.
        """
        return self.datasets.cluster.merge(
            dataset_id=self.dataset_id,
            vector_fields=self.vector_fields,
            alias=self.alias,
            cluster_ids=[target_cluster_id] + cluster_ids,
        )

    def create_byo_clusters(self):
        """
        Create BYO clusters for a given field
        """
        # TODO: Change into generator to make unique values more than 9999
        results = self.datasets.facets(
            dataset_id=self.dataset_id, fields=[self.byo_cluster_field], page_size=9999
        )

        try:
            for r in results["results"][self.byo_cluster_field]:
                filters = [
                    {
                        "field": self.byo_cluster_field,
                        "filter_type": "exact_match",
                        "condition": "==",
                        "condition_value": r["value"],
                    }
                ]
                cluster_doc = {}
                self.set_field(self._get_cluster_field_name(), cluster_doc, r["value"])
                results = self.datasets.documents.update_where(
                    dataset_id=self.dataset_id, update=cluster_doc, filters=filters
                )

        except KeyError:
            raise ValueError("Cluster field has no values.")

        return results

    def _operate(self, cluster_id: str, field: str, output: dict, func: Callable):
        """
        Internal function for operations

        It takes a cluster_id, a field, an output dictionary, and a function, and then it gets all the
        documents in the cluster, gets the field across all the documents, and then applies the function
        to the field

        Parameters
        ----------
        cluster_id : str
            str, field: str, output: dict, func: Callable
        field : str
            the field you want to get the value for
        output : dict
            dict
        func : Callable
            Callable

        """
        cluster_field = self._get_cluster_field_name()
        # TODO; change this to fetch all documents
        documents = self._get_all_documents(
            self.dataset_id,
            filters=[
                {
                    "field": cluster_field,
                    "filter_type": "exact_match",
                    "condition": "==",
                    "condition_value": cluster_id,
                },
                {
                    "field": field,
                    "filter_type": "exists",
                    "condition": ">=",
                    "condition_value": " ",
                },
            ],
            select_fields=[field, cluster_field],
            show_progress_bar=False,
        )
        # get the field across each
        arr = self.get_field_across_documents(field, documents)
        output[cluster_id] = func(arr)

    def _operate_across_clusters(self, field: str, func: Callable):
        output: Dict[str, Any] = dict()
        for cluster_id in self.list_cluster_ids():
            self._operate(cluster_id=cluster_id, field=field, output=output, func=func)
        return output

    @property
    def labels(self):
        metadata = self.datasets.metadata(self.dataset_id)
        metadata = metadata.get("results", {}).get("cluster_metadata", {})
        vector_fields = ["text_mpnet_vector_"]
        alias = "kmeans-100"

        cluster_field = "_cluster_." + ".".join(vector_fields) + "." + alias
        labels = metadata.get("labels", {}).get(cluster_field, {})
        print("To view nicely, please use `pd.DataFrame(labels)`.")
        return labels

    def create_parent_cluster(
        self,
        to_merge: dict,
        new_cluster_field: str,
    ):
        """
        to_merge should look similar to below:

        .. code-block::

            to_merge = {
                0: [
                    'cluster_1',
                    'cluster_2'
                ]
            }

        """
        parent_field = self._get_cluster_field_name()
        for cluster, clusters_to_combine in to_merge.items():
            for i, cluster_to_combine in enumerate(clusters_to_combine):
                # Update the documents in the cluster to become a subcluster
                cluster_filter = [
                    {
                        "field": parent_field,
                        "filter_type": "contains",
                        "condition": "==",
                        "condition_value": cluster_to_combine,
                    }
                ]

                if "cluster_" in cluster_to_combine:
                    cluster_to_combine = cluster_to_combine.replace("cluster_", "")
                if isinstance(cluster, int):
                    update = {
                        new_cluster_field: f"mergedCluster_{cluster}-{cluster_to_combine}"
                    }
                elif isinstance(cluster, str):
                    update = {new_cluster_field: f"{cluster}-{cluster_to_combine}"}
                updated = self.datasets.documents.update_where(
                    dataset_id=self.dataset_id, update=update, filters=cluster_filter
                )
                print("Update status: ")
                print(updated)

            # Merge the original clusters combine to create a subcluster
            if isinstance(cluster, int):
                try:
                    merge_results = self.merge(
                        target_cluster_id=clusters_to_combine[0],
                        # target_cluster_id=f"mergedCluster_{cluster}",
                        cluster_ids=clusters_to_combine[1:],
                    )
                    print(merge_results)
                except Exception as e:
                    print(e)
            elif isinstance(cluster, str):
                self.merge(target_cluster_id=cluster, cluster_ids=clusters_to_combine)

        self.append_metadata_list(
            field="_subcluster_",
            value_to_append={
                "parent_field": parent_field,
                "cluster_field": new_cluster_field,
            },
            only_unique=True,
        )

        # Port over the labels from the cluster to the subcluster
        metadata = self.datasets.metadata(self.dataset_id)["results"]
        if parent_field in metadata["cluster_metadata"]["labels"]:
            for k, old_labels in to_merge.items():
                for l in old_labels:
                    label = metadata["cluster_metadata"]["labels"][parent_field][
                        "labels"
                    ][l]
                    if new_cluster_field not in metadata["cluster_metadata"]["labels"]:
                        metadata["cluster_metadata"]["labels"].update(
                            {new_cluster_field: {"labels": {}}}
                        )
                    l = l.replace("cluster_", "")
                    cluster_id = f"mergedCluster_{k}-{l}"
                    metadata["cluster_metadata"]["labels"][new_cluster_field]["labels"][
                        cluster_id
                    ] = label
            results = self.datasets.post_metadata(self.dataset_id, metadata)
            print("Updated metadata")
            print(results)

    def explain_text_clusters(
        self,
        text_field,
        encode_fn_or_model,
        n_closest: int = 5,
        highlight_output_field="_explain_",
        algorithm: str = "relational",
        model_kwargs: Optional[dict] = None,
    ):
        """
        It takes a text field and a function that encodes the text field into a vector.
        It then returns the top n closest vectors to each cluster centroid.
        .. code-block::
            def encode(X):
                return [1, 2, 1]
            cluster_ops.explain_text_clusters(text_field="hey", encode_fn_or_model=encode)

        Parameters
        ----------
        text_field
            The field in the dataset that contains the text to be explained.
        encode_fn
            This is the function that will be used to encode the text.
        n_closest : int, optional
            The number of closest documents to each cluster to return.
        highlight_output_field, optional
            The name of the field that will be added to the output dataset.
        algorithm: str
            Algorithm is either "centroid" or "relational"

        Returns
        -------
            A new dataset with the same data as the original dataset, but with a new field called _explain_
        """
        if isinstance(encode_fn_or_model, str):
            # Get the model
            from relevanceai.operations_new.vectorize.text.transform import (
                VectorizeTextTransform,
            )

            self.model = VectorizeTextTransform._get_model(encode_fn_or_model)
            encode_fn = self.model.encode
        else:
            encode_fn = encode_fn_or_model

        from relevanceai.operations_new.cluster.text.explainer.ops import (
            TextClusterExplainerOps,
        )

        ops = TextClusterExplainerOps(credentials=self.credentials)
        if algorithm == "centroid":
            return ops.explain_clusters(
                dataset_id=self.dataset_id,
                alias=self.alias,
                vector_fields=self.vector_fields,
                text_field=text_field,
                encode_fn=encode_fn,
                n_closest=n_closest,
                highlight_output_field=highlight_output_field,
            )
        elif algorithm == "relational":
            return ops.explain_clusters_relational(
                dataset_id=self.dataset_id,
                alias=self.alias,
                vector_fields=self.vector_fields,
                text_field=text_field,
                encode_fn=encode_fn,
                n_closest=n_closest,
                highlight_output_field=highlight_output_field,
            )
        raise ValueError("Algorithm needs to be either `relational` or `centroid`.")

    @track
    def aggregate(
        self,
        metrics: Optional[list] = None,
        sort: Optional[list] = None,
        groupby: Optional[list] = None,
        filters: Optional[list] = None,
        page_size: int = 20,
        page: int = 1,
        asc: bool = False,
        flatten: bool = True,
    ):
        """
        Takes an aggregation query and gets the aggregate of each cluster in a collection. This helps you interpret each cluster and what is in them.
        It can only can be used after a vector field has been clustered. \n
        Aggregation/Groupby of a collection using an aggregation query. The aggregation query is a json body that follows the schema of:

        .. code-block::

            {
                "groupby" : [
                    {"name": <alias>, "field": <field in the collection>, "agg": "category"},
                    {"name": <alias>, "field": <another groupby field in the collection>, "agg": "numeric"}
                ],
                "metrics" : [
                    {"name": <alias>, "field": <numeric field in the collection>, "agg": "avg"}
                    {"name": <alias>, "field": <another numeric field in the collection>, "agg": "max"}
                ]
            }

        For example, one can use the following aggregations to group score based on region and player name.

        .. code-block::

            {
                "groupby" : [
                    {"name": "region", "field": "player_region", "agg": "category"},
                    {"name": "player_name", "field": "name", "agg": "category"}
                ],
                "metrics" : [
                    {"name": "average_score", "field": "final_score", "agg": "avg"},
                    {"name": "max_score", "field": "final_score", "agg": "max"},
                    {'name':'total_score','field':"final_score", 'agg':'sum'},
                    {'name':'average_deaths','field':"final_deaths", 'agg':'avg'},
                    {'name':'highest_deaths','field':"final_deaths", 'agg':'max'},
                ]
            }
        "groupby" is the fields you want to split the data into. These are the available groupby types:
            - category : groupby a field that is a category
            - numeric: groupby a field that is a numeric
        "metrics" is the fields and metrics you want to calculate in each of those, every aggregation includes a frequency metric. These are the available metric types:
            - "avg", "max", "min", "sum", "cardinality"
        The response returned has the following in descending order. \n
        If you want to return documents, specify a "group_size" parameter and a "select_fields" parameter if you want to limit the specific fields chosen. This looks as such:
            .. code-block::
                {
                    'groupby':[
                        {'name':'Manufacturer','field':'manufacturer','agg':'category',
                        'group_size': 10, 'select_fields': ["name"]},
                    ],
                    'metrics':[
                        {'name':'Price Average','field':'price','agg':'avg'},
                    ],
                }
                # ouptut example:
                {"title": {"title": "books", "frequency": 200, "documents": [{...}, {...}]}, {"title": "books", "frequency": 100, "documents": [{...}, {...}]}}
        For array-aggregations, you can add "agg": "array" into the aggregation query.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        metrics: list
            Fields and metrics you want to calculate
        groupby: list
            Fields you want to split the data into
        filters: list
            Query for filtering the search results
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        asc: bool
            Whether to sort results by ascending or descending order
        flatten: bool
            Whether to flatten
        alias: string
            Alias used to name a vector field. Belongs in field_{alias} vector
        metrics: list
            Fields and metrics you want to calculate
        groupby: list
            Fields you want to split the data into
        filters: list
            Query for filtering the search results
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        asc: bool
            Whether to sort results by ascending or descending order
        flatten: bool
            Whether to flatten

        Example
        ---------
        .. code-block::
            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset_id")
            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)
            cluster_ops.run(df, vector_fields=["sample_vector_"])
            clusterer.aggregate(
                "sample_dataset_id",
                groupby=[{
                    "field": "title",
                    "agg": "wordcloud",
                }],
                vector_fields=['sample_vector_']
            )
        """
        metrics = [] if metrics is None else metrics
        groupby = [] if groupby is None else groupby
        filters = [] if filters is None else filters
        sort = [] if sort is None else sort

        return self.datasets.cluster.aggregate(
            dataset_id=self.dataset_id,
            vector_fields=self.vector_fields,
            groupby=groupby,
            metrics=metrics,
            sort=sort,
            filters=filters,
            alias=self.alias,
            page_size=page_size,
            page=page,
            asc=asc,
            flatten=flatten,
        )
